import subprocess
import sys
from math import pi
from distutils.spawn import find_executable
import pandas as pd
from va.utils.misc import *


def run_phenixmmfsc(full_modelpath, full_mappath, output_path):
    """

        Phenix score

    :return:
    """

    errlist = []
    try:
        assert find_executable('phenix.mtriage') is not None
        phenixpath = find_executable('phenix.mtriage')
        create_directory(output_path)
        phenixmmfsc_cmd = '{} {} {} fsc_curve_model=True d_fsc_model_05=True'.format(phenixpath, full_mappath, full_modelpath)
        print(phenixmmfsc_cmd)
        try:
            process = subprocess.Popen(phenixmmfsc_cmd.split(), stdout=subprocess.PIPE, stderr=subprocess.STDOUT, cwd=output_path)
            output = process.communicate('n\n')[0]
            errqscore = 'error'
            if sys.version_info[0] >= 3:
                for item in output.decode('utf-8').split('\n'):
                    # item = cline.decode('utf-8').strip()
                    print(item)
                    if errqscore in item.lower():
                        errline = item.strip()
                        errlist.append(errline)
                        assert errqscore not in output.decode('utf-8'), errline

            else:
                for item in output.split('\n'):
                    print(item)
                    if errqscore in item.lower():
                        errline = item.strip()
                        errlist.append(errline)
                        assert errqscore not in output.decode('utf-8'), errline
        except subprocess.CalledProcessError as suberr:
            err = 'Phenix map-model FSC calculation error: {}.'.format(suberr)
            errlist.append(err)
            sys.stderr.write(err + '\n')
    except AssertionError as exerr:
        err = 'Phenix executable is not there.'
        errlist.append(err)
        sys.stderr.write('Phenix executable is not there: {}\n'.format(exerr))

    return errlist

def read_mmfsc(mmfsc_log, errlist):
    """
        Read the residue-wise CCC from the output of Phenix results
    :param: cc_per_residue_log input file cc_per_residue.log file(full path name) from Phenix CC calculation
    :return:
    """

    if not errlist and os.path.isfile(mmfsc_log):
        df = pd.read_csv(mmfsc_log, delim_whitespace=True, header=None)
    else:
        df = None

    return df


def data_extra(data_extra):
    """
    Add half bit, one-bit and 3-sigma to data_fsc_block
    """

    frequencies = data_extra['level']
    asym = 1.0
    half_bit = []
    one_bit = []
    for i in range(0, len(frequencies)):
        volume_diff = (4.0 / 3.0) * pi * ((i + 1) ** 3 - i ** 3)
        novox_ring = volume_diff / (1 ** 3)
        effno_vox = (novox_ring * ((1.5 * 0.66) ** 2)) / (2 * asym)
        if effno_vox < 1.0: effno_vox = 1.0
        sqreffno_vox = np.sqrt(effno_vox)

        bit_value = (0.2071 + 1.9102 / sqreffno_vox) / (1.2071 + 0.9102 / sqreffno_vox)
        half_bit.append(keep_three_significant_digits(bit_value))
        onebit_value = (0.5 + 2.4142 / sqreffno_vox) / (1.5 + 1.4142 / sqreffno_vox)
        one_bit.append(keep_three_significant_digits(onebit_value))

    gold_line = [0.143] * len(frequencies)
    half_line = [0.5] * len(frequencies)
    half_bit.insert(0, 1)
    one_bit.insert(0, 1)

    data_extra['halfbit'] = half_bit[:-1]
    data_extra['onebit'] = one_bit[:-1]
    data_extra['0.5'] = half_line
    data_extra['0.143'] = gold_line

    return data_extra

def _xy_check(x, y):
    """
    check the x, y value and return the results
    """

    if x.size == 0 and y.size == 0:
        return None, None
    else:
        x = np.round(x[0][0], 4)
        y = np.round(y[0][0], 4)
        return x, y
def all_intersection(allcurves):
    """
    Get all intersections from data_fsc_block
    """

    mmfsc = allcurves['level']
    correlation = allcurves['fsc']
    half_bit = allcurves['halfbit']
    gold = allcurves['0.143']
    half = allcurves['0.5']

    x_gold, y_gold = interpolated_intercept(mmfsc, correlation, gold)
    x_half, y_half = interpolated_intercept(mmfsc, correlation, half)
    x_half_bit, y_half_bit = interpolated_intercept(mmfsc, correlation, half_bit)

    x_gold, y_gold = _xy_check(x_gold, y_gold)
    x_half, y_half = _xy_check(x_half, y_half)
    x_half_bit, y_half_bit = _xy_check(x_half_bit, y_half_bit)
    if not x_gold or not y_gold:
        print('!!! No intersection between FSC and 0.143 curves.')
    if not x_half or not y_half:
        print('!!! No intersection between FSC and 0.143 curves.')
    if not x_half_bit or not y_half_bit:
        print('!!! No intersection between FSC and 0.143 curves.')

    intersections = {'halfbit': {'x': x_half_bit, 'y': y_half_bit},
                      '0.5': {'x': x_half, 'y': y_half},
                      '0.143': {'x': x_gold, 'y': y_gold}}

    return intersections

def mmfscdf_todict(mmfscdf, nyquist):
    """
        Given the Phenix CC result in dataframe and output the dict to be save in json
    :param ccdf: phenix cc_per_residue.log result into dataframe
    :return: dict which contain CC per residue results
    """

    real_mmfscdf = mmfscdf[mmfscdf[0] <= nyquist].round(3).drop_duplicates()
    skip = (len(real_mmfscdf) - 2) / (96 - 1)

    # Select the rows based on the calculated skip value
    indices = [0]  # Include the first row
    indices.extend(int(i * skip) + 1 for i in range(1, 95))  # Select evenly distributed rows
    indices.append(len(real_mmfscdf) - 1)  # Include the last row
    df_evenly = real_mmfscdf.iloc[indices]
    x = df_evenly.iloc[:, 0].tolist()
    y = df_evenly.iloc[:, 1].tolist()
    curves = {}
    curves['fsc'] = y
    curves['level'] = x
    allcurves = data_extra(curves)
    intersections = all_intersection(allcurves)

    finaldict = {'curves': curves, 'intersections': intersections}

    return finaldict




import os
import sys
from Bio.PDB import MMCIFParser
from Bio.PDB.mmcifio import MMCIFIO
from Bio.PDB.PDBIO import Select
# from Bio.PDB.MMCIF2Dict import MMCIF2Dict
from Bio.PDB.Structure import Structure
import timeit


class NotDisordered(Select):
    """
        Class used to select non-disordered atom from biopython structure instance
    """

    def accept_atom(self, atom):
        """
            Accept only atoms that at "A"
        :param atom: atom instance from biopython library
        :return: True or False
        """
        if (not atom.is_disordered()) or atom.get_altloc() == "A":
            atom.set_altloc(" ")
            return True
        else:
            return False


class Model:
    """
        MapProcessor class contains methods deal with map processing method and model associated map processing methods
        Instance can be initialized with either full map file path or a mrcfile map object
    """

    def __init__(self, input_model, input_dir=None):
        """
            Initialization method with input_model as full path or as a biopython mmcif object and the directory of it
        """
        self.model, self.model_dir = self._set_model_and_dir(input_model) if not input_model else self._set_model_and_dir(input_model, input_dir)

    @staticmethod
    def _set_model_and_dir(input_model, input_dir=None):
        if isinstance(input_model, str) and os.path.isfile(input_model):
            return input_model, os.path.dirname(input_model)
        elif isinstance(input_model, Structure) and input_dir:
            return input_model, input_dir
        else:
            return None, None
    def get_auth_comp_id_map(self, input_cif):
        '''
            Extract the auth_comp_id mapping from an mmCIF dictionary.
        Parameters:
        mmcif_dict (dict): Parsed mmCIF dictionary

        Returns:
        dict: A dictionary mapping (chain_id, resseq) to auth_comp_id
        '''

        parser = MMCIFParser()
        structure = parser.get_structure('structure', input_cif)
        mmcif_dict = parser._mmcif_dict
        chains = mmcif_dict['_atom_site.auth_asym_id']
        resseqs = mmcif_dict['_atom_site.label_seq_id']
        auth_comp_ids = mmcif_dict['_atom_site.auth_comp_id']

        auth_comp_id_map = {}
        for chain, resseq, auth_comp_id in zip(chains, resseqs, auth_comp_ids):
            key = (chain, int(resseq))
            auth_comp_id_map[key] = auth_comp_id

        return auth_comp_id_map

    def get_formal_charge_map(self, input_cif):
        '''
            Extract the pdbx_formal_charge mapping from an mmCIF dictionary.
        Parameters:
        mmcif_dict (dict): Parsed mmCIF dictionary

        Returns:
        dict: A dictionary mapping (chain_id, resseq) to pdbx_formal_charge
        '''

        parser = MMCIFParser()
        structure = parser.get_structure('structure', input_cif)
        mmcif_dict = parser._mmcif_dict
        chains = mmcif_dict['_atom_site.auth_asym_id']
        resseqs = mmcif_dict['_atom_site.label_seq_id']
        formal_charges = mmcif_dict['_atom_site.pdbx_formal_charge']

        formal_charge_map = {}
        for chain, resseq, formal_charge in zip(chains, resseqs, formal_charges):
            key = (chain, int(resseq))
            formal_charge_map[key] = formal_charge

        return formal_charge_map

    def save_updated_model(self, input_cif_file_one, input_cif_file):
        '''
            Save the first model from the mmCIF file to a new mmCIF file with auth_comp_id.
        Parameters:
        input_cif_file (str): Input mmCIF file path
        output_cif_file (str): Output mmCIF file path
        '''

        # Create the auth_comp_id_map
        auth_comp_id_map = self.get_auth_comp_id_map(input_cif_file_one)
        formal_charge_map = self.get_formal_charge_map(input_cif_file_one)

        # Extract the first model
        parser = MMCIFParser()
        structure = parser.get_structure('structure', input_cif_file)
        mmcif_dict = parser._mmcif_dict
        first_model = structure[0]

        # Prepare to add auth_comp_id
        if '_atom_site.auth_comp_id' not in mmcif_dict:
            mmcif_dict['_atom_site.auth_comp_id'] = []
        if '_atom_site.pdbx_formal_charge' not in mmcif_dict:
            mmcif_dict['_atom_site.pdbx_formal_charge'] = []

        # Get the existing residue information
        residues = mmcif_dict['_atom_site.label_comp_id']
        chains = mmcif_dict['_atom_site.auth_asym_id']
        resseqs = mmcif_dict['_atom_site.label_seq_id']

        # Add auth_comp_id based on the map
        for res, chain, resseq in zip(residues, chains, resseqs):
            auth_comp_id = auth_comp_id_map.get((chain, int(resseq)), res)
            formal_charge = formal_charge_map.get((chain, int(resseq)), '?')
            mmcif_dict['_atom_site.auth_comp_id'].append(auth_comp_id)
            mmcif_dict['_atom_site.pdbx_formal_charge'].append(formal_charge)

        # Write the modified dictionary to a new mmCIF file
        io = MMCIFIO()
        io.set_dict(mmcif_dict)
        io.save(input_cif_file)

    def _structure_to_model(self, pid, cur_model_name):
        """
            Take structure object and output the used model object

        :param pid: String of anything or pdbid
        :param cur_model_name: String of the model name
        :return: TEMPy model instance which will be used for the whole calculation
        """

        p = MMCIFParser()
        io = MMCIFIO()
        org_file_name = cur_model_name
        match_text = 'data_'
        out_moderate_cif = cur_model_name + '_moderated.cif'
        match = self.remove_lines_after_match(cur_model_name, out_moderate_cif, match_text)
        if match:
            structure = p.get_structure(pid, out_moderate_cif)
            cur_model_name = out_moderate_cif
        else:
            structure = p.get_structure(pid, cur_model_name)

        if len(structure.get_list()) > 1:
            org_model = cur_model_name + '_org.cif'
            os.rename(cur_model_name, org_model)
            fstructure = structure[0]
            io.set_structure(fstructure)
            io.save(cur_model_name)
            self.save_updated_model(org_file_name, cur_model_name)
            used_frame = p.get_structure('first', cur_model_name)
            print('!!!There are multiple models in the cif file. Here we only use the first for calculation.')
        else:
            used_frame = structure

        # io.set_structure(used_frame)
        if self.has_disorder_atom(used_frame):
            cur_model_name = cur_model_name + '_Alt_A.cif'
            io.set_structure(used_frame)
            print('There are alternative atom in the model here we only use A for calculations and saved as {}'
                  .format(cur_model_name))
            io.save(cur_model_name, select=NotDisordered())
            new_structure = p.get_structure(pid, cur_model_name)
        else:
            new_structure = used_frame

        setattr(new_structure, "filename", org_file_name)
        tmodel = new_structure

        return tmodel

    @staticmethod
    def remove_lines_after_match(input_file, output_file, match_text):
        """
            Remove lines after the match text
        """

        try:
            with open(input_file, 'r') as infile:
                found_second_match = False
                match_count = 0
                content_before_second_match = []

                for line in infile:
                    if line.startswith(match_text):
                        match_count += 1
                        if match_count >= 2:
                            found_second_match = True
                            break
                    content_before_second_match.append(line)

            if found_second_match:
                with open(output_file, 'w') as outfile:
                    outfile.writelines(content_before_second_match)
                return True
            else:
                print('Second match not found in the file.')
                return False

        except FileNotFoundError:
            print(f'Error: Input file "{input_file}" not found.')
            return False
        except Exception as e:
            print(f'An error occurred: {str(e)}')

            return False

    @staticmethod
    def has_disorder_atom(structure):
        """
            Check if the model contains disorder atoms
        """

        ress = structure.get_residues()
        disorder_flag = False
        for res in ress:
            if res.is_disordered() == 1:
                disorder_flag = True
                return disorder_flag
        return disorder_flag

    def read_model_test(self):
        """
            Read models if '-f' argument is used
        """

        start = timeit.default_timer()
        if self.model is not None:
            model_name = self.model
            full_model_name = model_name
            try:
                model_size = os.stat(full_model_name).st_size
                pid = 'id'
                input_model = []
                tmodel = self._structure_to_model(pid, full_model_name)
                input_model.append(tmodel)

                end = timeit.default_timer()
                print('Read model time: %s' % (end - start))
                print('------------------------------------')

                return input_model, pid, model_size
            except:
                print('!!! File: %s does not exist or corrupted: %s!!!' % (full_model_name, sys.exc_info()[1]))
                print('------------------------------------')
                input_model = None

                return input_model
        else:
            print('No model is given.')
            input_model = None

            return input_model


    def final_model(self):
        """
            Get the final biopython model object
        """
        if isinstance(self.model, Structure):
            return self.model
        else:
            return self.read_model_test()




import os
import sys
import timeit
import json
import codecs
import cv2
import numpy as np
from math import ceil
from scipy import ndimage
from mrcfile.mrcfile import MrcFile
import inspect
from va.utils.misc import out_json

class Projections:
    def __init__(self, primary_map, rawmap=None, workdir=None, platform=None):
        self.map = primary_map
        self.rawmap = rawmap
        self.workdir = workdir
        self.platform = platform
        self.errlist = []

    def mapincheck(self, mapin, workdirin):
        frame = inspect.currentframe()
        func = inspect.getframeinfo(frame.f_back).function
        map, workdir = None, None

        if mapin is not None:
            if isinstance(mapin, str):
                if os.path.isfile(mapin):
                    map = MrcFile(mapin, mode='r')
                else:
                    print('Map does not exist.')
            elif isinstance(mapin, MrcFile):
                map = mapin
            else:
                print(f'Function:{func} only accepts a string of the full map name or a TEMPy Map object as input.')

        if workdirin is not None:
            if isinstance(workdirin, str):
                if os.path.isdir(workdirin):
                    workdir = workdirin
                else:
                    print('Output directory does not exist.')
            else:
                print(f'Function:{func} only accepts a string as the directory parameter.')

        return map, workdir

    def resize_img(self, input_arr, output_img):
        height, width = input_arr.shape[0:2]
        if width >= height:
            largerscaler = 300. / width
            newheight = int(ceil(largerscaler * height))
            resized_img = cv2.resize(input_arr, (300, newheight), interpolation=cv2.INTER_LANCZOS4)
        else:
            largerscaler = 300. / height
            newwidth = int(ceil(largerscaler * width))
            resized_img = cv2.resize(input_arr, (newwidth, 300), interpolation=cv2.INTER_LANCZOS4)

        cv2.imwrite(output_img, resized_img)

    def scale_img(self, projection):
        onevalue = True if projection.min() == projection.max() else False
        if onevalue:
            rescaled = np.full((projection.shape), 0.0).astype('uint8')
        else:
            rescaled = (((projection - projection.min()) * 255.0) / (projection.max() - projection.min())).astype('uint8')
        return rescaled

    def get_projection(self, map, type, axis):
        mapdata = map.data
        ind = None
        if type == 'projection':
            projection = np.sum(mapdata, axis=axis)
        elif type == 'max':
            projection = np.amax(mapdata, axis=axis)
        elif type == 'min':
            projection = np.amin(mapdata, axis=axis)
        elif type == 'std':
            projection = np.std(mapdata, axis=axis)
        elif type == 'median':
            projection = np.median(mapdata, axis=axis)
        elif type == 'central':
            projection = map.data[:, :, int(map.header.nx / 2)] if axis == 2 else \
                         map.data[:, int(map.header.ny / 2), :] if axis == 1 else \
                         map.data[int(map.header.nz / 2), :, :] if axis == 0 else None
            if axis == 2:
                ind = {'x': int(map.header.nx / 2)}
            elif axis == 1:
                ind = {'y': int(map.header.ny / 2)}
            elif axis == 0:
                ind = {'z': int(map.header.nz / 2)}
            else:
                ind = None

        elif type == 'largestvariance':
            if axis == 2:
                xvar = [ndimage.variance(map.data[:, :, i]) for i in range(map.data.shape[2])]
                xlargeind = int(np.argmax(xvar))
                ind = {'x': xlargeind}
                projection = map.data[:, :, xlargeind]
            elif axis == 1:
                yvar = [ndimage.variance(map.data[:, i, :]) for i in range(map.data.shape[1])]
                ylargeind = int(np.argmax(yvar))
                ind = {'y': ylargeind}
                projection = map.data[:, ylargeind, :]
            elif axis == 0:
                zvar = [ndimage.variance(map.data[i, :, :]) for i in range(map.data.shape[0])]
                zlargeind = int(np.argmax(zvar))
                ind = {'z': zlargeind}
                projection = map.data[zlargeind, :, :]
            else:
                projection = None
        else:
            projection = None
            print('Wrong type. Use one of: projection, max, min, std, median, central, largestvariance.')
        return projection, ind


    def output_filename(self, axis_letter, mapname, type, proc_fun=None):
        if proc_fun:
            suffix = 'glow_' if proc_fun == self.glowimage else ''
        else:
            suffix = ''

        if type == 'central' or type == 'largestvariance':
            output_name = f'{self.workdir}{mapname}_{axis_letter}{type}_slice.jpeg'
            scaled_output_name = f'{self.workdir}{mapname}_scaled_{axis_letter}{type}_slice.jpeg'
        else:
            output_name = f'{self.workdir}{mapname}_{suffix}{axis_letter}{type}.jpeg'
            scaled_output_name = f'{self.workdir}{mapname}_scaled_{suffix}{axis_letter}{type}.jpeg'

        return output_name, scaled_output_name

    def save_scale(self, reposition, output_name, scaled_output_name, errlist):

        try:
            cv2.imwrite(output_name, reposition)
        except IOError as ioerr:
            err = f'Saving {output_name} projection err: {ioerr}'
            errlist.append(err)
            sys.stderr.write(err + '\n')

        try:
            self.resize_img(reposition, scaled_output_name)
        except Exception as e:
            err = f'Saving scaled {scaled_output_name} image err: {e}'
            errlist.append(err)
            sys.stderr.write(err + '\n')

    def map_to_img(self, map, axis, type, errlist, proc_fun=None):
        projection, ind = self.get_projection(map, type, axis)
        org = {}
        scale = {}
        if projection is not None:
            rescaled = self.scale_img(projection)
            if axis == 2:
                reposition = np.flipud(rescaled)
                axis_letter = 'x'
            elif axis == 1:
                reposition = np.rot90(rescaled)
                axis_letter = 'y'
            elif axis == 0:
                reposition = np.flipud(rescaled)
                axis_letter = 'z'
            else:
                return None

            mapname = os.path.basename(map.fullname)
            output_name, scaled_output_name = self.output_filename(axis_letter, mapname, type, proc_fun)
            org[axis_letter] = os.path.basename(output_name)
            scale[axis_letter] = os.path.basename(scaled_output_name)

            self.save_scale(reposition, output_name, scaled_output_name, errlist)
            if proc_fun:
                glowimage = proc_fun(reposition)
                self.save_scale(glowimage, output_name, scaled_output_name, errlist)

        return ind, org, scale

    def orthogonal_projections(self, mapin=None, workdir=None, type=None, label=''):
        map, workdir = self.mapincheck(mapin, workdir)
        if map is not None and workdir is not None:
            start = timeit.default_timer()
            self.errlist = []
            mapname = os.path.basename(map.fullname)
            if not type:
                types = ['projection', 'max', 'min', 'std', 'median', 'central', 'largestvariance']
            else:
                types = [type]
            result_dict = {}
            for type in types:
                ind_result = {}
                org_result = {}
                scale_result = {}
                final_org = {}
                final_scale = {}
                final_ind = {}
                glow_ind_result = {}
                glow_org_result = {}
                glow_scale_result = {}
                for axis in range(2, -1, -1):
                    ind, org, scale = self.map_to_img(map, axis, type, self.errlist)
                    org_result.update(org)
                    scale_result.update(scale)
                    if type == 'central' or type == 'largestvariance':
                        ind_result.update(ind)
                    if type == 'max' or type == 'projection' or type == 'std':
                        glow_ind, glow_org, glow_scale = self.map_to_img(map, axis, type, self.errlist, self.glowimage)
                        glow_org_result.update(glow_org)
                        glow_scale_result.update(glow_scale)
                        # if type == 'central' or type == 'largestvariance':
                        #     glow_ind_result.update(glow_ind)
                    # if type == 'central' or type == 'largestvariance':
                    #     ind_result.update(ind)
                final_org['original'] = org_result
                final_scale['scaled'] = scale_result
                if type == 'central' or type == 'largestvariance':
                    final_ind['indices'] = ind_result

                if type == 'largestvariance':
                    result_dict['largest_variance_slice'] = {**final_org, **final_scale, **final_ind}
                elif type == 'central':
                    result_dict['central_slice'] = {**final_org, **final_scale, **final_ind}
                else:
                    result_dict[f'orthogonal_{type}'] = {**final_org, **final_scale}

            if len(types) == 1:
                json_file = f'{self.workdir}{mapname}_{type}.json'
            else:
                if not label:
                    json_file = f'{self.workdir}{mapname}_primary_projections.json'
                else:
                    json_file = f'{self.workdir}{mapname}_raw_projections.json'

            out_json(result_dict, json_file)
            end = timeit.default_timer()
            print(f'Projections and their glow projections time for {mapname}: {end - start}')
            print('------------------------------------')
        else:
            print('No orthogonal projections without proper map input and the output directory information.')

    def rawmap_projections(self):
        if self.rawmap is not None:
            self.orthogonal_projections(self.rawmap, self.workdir, 'rawmap_')
        else:
            print('No raw map to calculate orthogonal projections.')


    def save_to_json(self, errlist, mapname, workdir, label):
        bothjson = {}
        projectionjson = {}
        orgprojectionjson = {}

        for axis_letter in ['x', 'y', 'z']:
            projectionjson[axis_letter] = {}
            orgprojectionjson[axis_letter] = {}

        for axis in range(3):
            for type in ['projection', 'max', 'min', 'std', 'median', 'central', 'largestvariance']:
                for axis_letter in ['x', 'y', 'z']:
                    projection_name = f'{workdir}{mapname}_{axis_letter}{type}.jpeg'
                    scaled_projection_name = f'{workdir}{mapname}_scaled_{axis_letter}{type}.jpeg'
                    projectionjson[axis_letter][type] = os.path.basename(projection_name) if os.path.isfile(projection_name) else None
                    orgprojectionjson[axis_letter][type] = os.path.basename(scaled_projection_name) if os.path.isfile(scaled_projection_name) else None

        bothjson['original'] = orgprojectionjson
        bothjson['scaled'] = projectionjson

        if errlist:
            bothjson['err'] = {'orthogonal_median_err': errlist}
        finaldict = {label + 'orthogonal_median': bothjson}

        try:
            with codecs.open(f'{workdir}{mapname}_median.json', 'w', encoding='utf-8') as f:
                json.dump(finaldict, f)
        except IOError as ioerr:
            jsonerr = f'Saving median projection into json error: {ioerr}'
            sys.stderr.write(jsonerr + '\n')

    def glowimage(self, im_gray):
        """Applies a glow color map using cv2.applyColorMap()"""

        # Create the LUT:
        lut = np.zeros((256, 1, 3), dtype=np.uint8)
        lut[:, 0, 2] = [0, 1, 1, 2, 2, 3, 4, 6, 8, 11, 13, 15, 18,
                        21, 23, 24, 27, 30, 31, 33, 36, 37, 40, 42, 45, 46,
                        49, 51, 53, 56, 58, 60, 62, 65, 68, 70, 72, 74, 78,
                        80, 82, 84, 86, 89, 91, 93, 96, 98, 100, 102, 104, 106,
                        108, 110, 113, 115, 117, 119, 122, 125, 127, 129, 132, 135, 135,
                        137, 140, 141, 142, 145, 148, 149, 152, 154, 156, 157, 158, 160,
                        162, 164, 166, 168, 170, 171, 173, 174, 176, 178, 179, 180, 182,
                        183, 185, 186, 189, 192, 193, 193, 194, 195, 195, 196, 198, 199,
                        201, 203, 204, 204, 205, 206, 207, 209, 211, 211, 211, 211, 213,
                        215, 216, 216, 216, 216, 218, 219, 219, 219, 220, 222, 223, 223,
                        223, 223, 224, 224, 226, 227, 227, 227, 227, 228, 229, 231, 231,
                        231, 231, 231, 231, 231, 232, 233, 234, 234, 234, 234, 234, 234,
                        235, 237, 238, 238, 238, 238, 238, 238, 238, 238, 239, 240, 242,
                        242, 242, 242, 242, 242, 242, 242, 242, 243, 245, 246, 246, 246,
                        246, 245, 245, 245, 245, 245, 245, 245, 247, 248, 249, 249, 249,
                        249, 249, 249, 249, 249, 249, 249, 249, 249, 249, 249, 249, 249,
                        249, 250, 252, 253, 253, 253, 253, 253, 253, 253, 253, 253, 253,
                        253, 253, 253, 253, 253, 253, 253, 253, 253, 253, 253, 253, 253,
                        253, 253, 253, 253, 253, 253, 253, 253, 253, 253, 253, 253, 253,
                        253, 253, 253, 253, 253, 253, 253, 253, 0]

        lut[:, 0, 1] = [138, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                        1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2,
                        2, 2, 2, 2, 3, 3, 3, 3, 4, 5, 5, 5, 5,
                        6, 6, 6, 7, 7, 8, 9, 9, 9, 9, 10, 10, 11,
                        12, 13, 14, 14, 14, 14, 15, 16, 16, 17, 18, 19, 19,
                        19, 20, 21, 22, 23, 24, 24, 25, 27, 28, 28, 28, 29,
                        31, 32, 32, 33, 35, 36, 36, 37, 39, 40, 40, 41, 42,
                        43, 45, 46, 47, 48, 50, 51, 52, 54, 55, 56, 57, 59,
                        62, 63, 65, 66, 68, 70, 71, 73, 74, 75, 77, 79, 81,
                        83, 85, 87, 89, 92, 93, 96, 98, 99, 100, 103, 105, 107,
                        109, 111, 113, 116, 117, 119, 121, 123, 125, 126, 128, 130, 132,
                        135, 137, 139, 140, 142, 144, 145, 147, 148, 150, 152, 154, 156,
                        158, 160, 161, 162, 164, 166, 168, 171, 172, 172, 174, 176, 177,
                        179, 180, 182, 183, 185, 187, 188, 189, 191, 192, 192, 192, 196,
                        200, 201, 201, 202, 203, 204, 206, 208, 210, 212, 213, 213, 214,
                        215, 217, 218, 220, 221, 222, 223, 224, 225, 226, 226, 227, 228,
                        229, 231, 232, 233, 234, 235, 236, 237, 238, 239, 239, 239, 239,
                        241, 242, 243, 243, 243, 244, 245, 247, 247, 247, 247, 247, 247,
                        248, 249, 250, 251, 251, 251, 252, 252, 252, 252, 252, 252, 252,
                        252, 252, 252, 252, 252, 252, 253, 253, 0]

        lut[:, 0, 0] = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                        1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 3, 4, 4,
                        4, 4, 4, 4, 3, 3, 3, 3, 4, 6, 6, 6, 6,
                        6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6,
                        6, 6, 6, 6, 6, 6, 6, 6, 5, 5, 7, 9, 10,
                        10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10,
                        10, 10, 10, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9,
                        9, 9, 9, 9, 8, 8, 8, 8, 8, 8, 8, 8, 8,
                        8, 8, 8, 8, 8, 8, 8, 8, 7, 7, 7, 7, 7,
                        7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 6, 6, 6,
                        6, 6, 6, 6, 7, 10, 12, 12, 12, 12, 12, 12, 13,
                        15, 17, 18, 18, 19, 20, 22, 23, 23, 24, 26, 27, 27,
                        28, 30, 32, 33, 34, 35, 37, 39, 40, 41, 43, 45, 46,
                        48, 50, 52, 54, 55, 57, 58, 61, 63, 65, 67, 70, 72,
                        73, 76, 79, 80, 83, 86, 88, 89, 92, 95, 96, 99, 102,
                        104, 107, 108, 110, 113, 116, 120, 121, 123, 125, 127, 129, 132,
                        135, 137, 140, 143, 146, 149, 150, 153, 155, 158, 160, 162, 165,
                        168, 171, 173, 175, 178, 180, 183, 186, 188, 191, 192, 195, 198,
                        201, 203, 206, 209, 210, 213, 215, 218, 221, 223, 225, 228, 231,
                        233, 236, 238, 241, 244, 246, 250, 252, 255]

        im_color = cv2.applyColorMap(im_gray, lut)
        return im_color

# Example usage:
# rawmap = MrcFile("/Users/zhe/Downloads/tests/VA/nfs/msd/work2/emdb/development/staging/em/81/8117/va/emd_8117_rawmap.map")
# workdir = "/Users/zhe/Downloads/tests/VA/nfs/msd/work2/emdb/development/staging/em/81/8117/va"
# op = Projections(rawmap, workdir)
# op.rawmap_projections()
# op.rawmap_median()



from math import pi
from va.utils.misc import *
import matplotlib.pyplot as plt


class GetStars:



    def __init__(self, star_file):
        self.star_file_name = os.path.basename(star_file)
        self.va_dir = os.path.dirname(os.path.dirname(os.path.dirname(star_file)))
        self.star_file = star_file
        self.star_blocks = self.read_star_file()


    def read_star_file(self):
        """
        find the data block which defined by 'block'
        """

        star_data = {}
        current_block = None

        with open(self.star_file, 'r') as file:
            for line in file:
                line = line.strip()
                if line.startswith("data_"):
                    current_block = line
                    star_data[current_block] = []
                elif current_block is not None and line:
                    star_data[current_block].append(line)

        return star_data


    def data_general_block(self):
        """
        Get the data_general block
        """

        data_general_block = {}
        for line in self.star_blocks['data_general']:
            if line.startswith('# version'):
                continue
            parts = line.split(maxsplit=1)
            if len(parts) == 2:
                key, value = parts
                data_general_block[key] = float(value.strip()) if value.isdigit() else value

        return data_general_block

    def randomise_from(self):
        """
            Get the resolution value that Relion is use for the phase randomization calculation
        """

        return float(self.data_general_block()['_rlnRandomiseFrom'])

    def data_fsc_block(self):
        """
        Get the data_fsc block
        """

        data_fsc_block = {}
        useful_data = self.star_blocks['data_fsc'][1:-2]
        header = [line.split(maxsplit=1)[0][4:] for line in useful_data[:8]]
        data = useful_data[8:]
        column_data_list = [row.split() for row in data]
        new_data_list = list(map(list, zip(*column_data_list)))
        # data_list = [[float(value) for value in sublist] for sublist in data_list]
        data_list = []
        for sublist in new_data_list:
            column_data_list = []
            for value in sublist:
                column_data_list.append(keep_three_significant_digits(float(value)))
            data_list.append(column_data_list)

        if len(header) == len(data_list):
            for header, data_list in zip(header, data_list):
                data_fsc_block[header] = data_list

        return data_fsc_block

    def data_extra(self, data_fsc_block):
        """
        Add half bit, one-bit and 3-sigma to data_fsc_block
        """

        indices = data_fsc_block['SpectralIndex']
        asym = 1.0
        half_bit = []
        one_bit = []
        for i in range(0, len(indices)):
            volume_diff = (4.0 / 3.0) * pi * ((i + 1) ** 3 - i ** 3)
            novox_ring = volume_diff / (1 ** 3)
            effno_vox = (novox_ring * ((1.5 * 0.66) ** 2)) / (2 * asym)
            if effno_vox < 1.0: effno_vox = 1.0
            sqreffno_vox = np.sqrt(effno_vox)

            bit_value = (0.2071 + 1.9102 / sqreffno_vox) / (1.2071 + 0.9102 / sqreffno_vox)
            half_bit.append(keep_three_significant_digits(bit_value))
            onebit_value = (0.5 + 2.4142 / sqreffno_vox) / (1.5 + 1.4142 / sqreffno_vox)
            one_bit.append(keep_three_significant_digits(onebit_value))

        gold_line = [0.143] * len(indices)
        half_line = [0.5] * len(indices)
        half_bit.insert(0, 1)
        one_bit.insert(0, 1)

        data_fsc_block['halfbit'] = half_bit[:-1]
        data_fsc_block['onebit'] = one_bit[:-1]
        data_fsc_block['0.5'] = half_line
        data_fsc_block['0.143'] = gold_line

        return data_fsc_block

    def _xy_check(self, x, y):
        """
        check the x, y value and return the results
        """

        if x.size == 0 and y.size == 0:
            return None, None
        else:
            x = np.round(x[0][0], 4)
            y = np.round(y[0][0], 4)
            return x, y

    def all_intersection(self, data_fsc_block):
        """
        Get all intersections from data_fsc_block
        """

        x = data_fsc_block['Resolution']
        correlation = data_fsc_block['FourierShellCorrelationUnmaskedMaps']
        half_bit = data_fsc_block['halfbit']
        gold = data_fsc_block['0.143']
        half = data_fsc_block['0.5']

        x_gold, y_gold = interpolated_intercept(x, correlation, gold)
        x_half, y_half = interpolated_intercept(x, correlation, half)
        x_half_bit, y_half_bit = interpolated_intercept(x, correlation, half_bit)

        x_gold, y_gold = self._xy_check(x_gold, y_gold)
        x_half, y_half = self._xy_check(x_half, y_half)
        x_half_bit, y_half_bit = self._xy_check(x_half_bit, y_half_bit)
        if not x_gold or not y_gold:
            print('!!! No intersection between FSC and 0.143 curves.')
        if not x_half or not y_half:
            print('!!! No intersection between FSC and 0.143 curves.')
        if not x_half_bit or not y_half_bit:
            print('!!! No intersection between FSC and 0.143 curves.')

        intersections = {'intersections': {
                          'halfbit': {'x': x_half_bit, 'y': y_half_bit},
                          '0.5': {'x': x_half, 'y': y_half},
                          '0.143': {'x': x_gold, 'y': y_gold}}}

        return intersections


    def all_curves(self, data_fsc_block):
        """
        Get all curves from data_fsc_block
        """

        curve_mapping = {
            'fsc': 'FourierShellCorrelationUnmaskedMaps',
            'onebit': 'onebit',
            'halfbit': 'halfbit',
            '0.5': '0.5',
            '0.143': '0.143',
            'level': 'Resolution',
            'angstrom_resolution': 'AngstromResolution',
            'phaserandmization': 'CorrectedFourierShellCorrelationPhaseRandomizedMaskedMaps',
            'phaserandomization': 'CorrectedFourierShellCorrelationPhaseRandomizedMaskedMaps',
            'fsc_masked': 'FourierShellCorrelationMaskedMaps',
            'fsc_corrected': 'FourierShellCorrelationCorrected'
        }
        curves = {key: data_fsc_block[value] for key, value in curve_mapping.items() if value in data_fsc_block}
        final_curves = {'curves': curves}

        return final_curves


    @staticmethod
    def intersections_into_curve(intersections, levels, corrected_curve, phase_rand_curve):
        """
            After getting the intersections and insert them into these curves
        :param intersections: list of data pairs containing intersections
        :param levels: list of x value of two intersect curves
        :param corrected_curve: list of value of correction curve correlation values
        :param phase_rand_curve: list of value of phase rand curve correlation
        """

        for a, b in intersections:

            exists = np.any(levels == a)
            if not exists:
                index = np.searchsorted(levels, a)
                levels = np.insert(levels, index, a)
                corrected_curve = np.insert(corrected_curve, index, b)
                phase_rand_curve = np.insert(phase_rand_curve, index, b)
            else:
                print(a, b)
                print('Intersection point on curve data')

        return levels, corrected_curve, phase_rand_curve

    def plot_fsc(self, data_curves, other_curves=None, other_curves_lable='Provided FSC'):
        """
            Plot Relion FSC curve
        :param intersections: list of tuples containing intersections in (x,y) coordinates
        :param levels: list of frequency values
        :param corrected_curve: list of corrected FSC value from Relion Star file
        :param phase_rand_curve: list of phase randomized FSC value from Relion
        """

        all_curves = None
        if 'curves' in data_curves.keys():
            all_curves = data_curves['curves']

        levels = None
        corrected_curve = None
        phase_rand_curve = None
        masked_curve = None
        unmasked_curve = None
        golden_line = None
        if all_curves:
            if 'level' in all_curves.keys():
                levels = np.array(all_curves['level'])
            if 'fsc_corrected' in all_curves.keys():
                corrected_curve = np.array(all_curves['fsc_corrected'])
            if 'phaserandmization' in all_curves.keys():
                phase_rand_curve = np.array(all_curves['phaserandmization'])
            if 'phaserandomization' in all_curves.keys():
                phase_rand_curve = np.array(all_curves['phaserandomization'])
            if 'fsc_masked' in all_curves.keys():
                masked_curve = np.array(all_curves['fsc_masked'])
            if 'fsc' in all_curves.keys():
                unmasked_curve = np.array(all_curves['fsc'])
            if '0.143' in all_curves.keys():
                golden_line = np.array(all_curves['0.143'])
            if other_curves:
                other_curve_x = other_curves[0]
                other_curve_y = other_curves[1]
                plt.plot(other_curve_x, other_curve_y, label=other_curves_lable)

            plt.plot(levels, unmasked_curve, label='FSC')
            plt.plot(levels, masked_curve, label='FSC (masked)')
            plt.plot(levels, corrected_curve, label='FSC (corrected)')
            plt.plot(levels, phase_rand_curve, label='Phaserandomization')
            plt.plot(levels, golden_line, linestyle='-.',  label='0.143')
            plt.legend(loc='upper right', fontsize='x-small')
            plt.savefig(self.va_dir + '/current_Relion_fsc.png')
            plt.close()

    def plot_feature_zone(self, intersections, levels, corrected_curve, phase_rand_curve):
        """
            Plot feature zone and save to image
        :param intersections: list of tuples containing intersections in (x,y) coordinates
        :param levels: list of frequency values
        :param corrected_curve: list of corrected FSC value from Relion Star file
        :param phase_rand_curve: list of phase randomized FSC value from Relion
        """

        frequency = 1 / self.randomise_from()
        plt.plot(levels, corrected_curve, color='red', label='t')
        plt.plot(levels, phase_rand_curve, color='blue', label='correlation')
        feature_area = 0
        for i in range(len(intersections)):
            if i == len(intersections) - 1:
                idx_start = np.where(levels == intersections[i][0])[0][0]
                idx_end = len(levels)
            else:
                idx_start = np.where(levels == intersections[i][0])[0][0]
                idx_end = np.where(levels == intersections[i + 1][0])[0][0]
            print(
                f'plt.fill_between(levels[{idx_start}:{idx_end}], corrected_curve[{idx_start}:{idx_end}], phase_rand_curve[{idx_start}:{idx_end}], where=(corrected_curve[{idx_start}:{idx_end}] >= phase_rand_curve[{idx_start}:{idx_end}]), color="gray", alpha=0.5)')
            plt.fill_between(levels[idx_start:idx_end], corrected_curve[idx_start:idx_end], phase_rand_curve[idx_start:idx_end],
                             where=(corrected_curve[idx_start:idx_end] >= phase_rand_curve[idx_start:idx_end]), color='pink', alpha=0.5)

            status = all(x >= y for x, y in zip(corrected_curve[idx_start:idx_end], phase_rand_curve[idx_start:idx_end]))

            if status:
                cur_area = np.trapz(corrected_curve[idx_start:idx_end] - phase_rand_curve[idx_start:idx_end], levels[idx_start:idx_end])
                feature_area += cur_area
                # validation area
                # area_one = np.trapz(corrected_curve[idx_start:idx_end], levels[idx_start:idx_end])
                # area_two = np.trapz(phase_rand_curve[idx_start:idx_end], levels[idx_start:idx_end])
                # print(area_one - area_two)
                # print(cur_area)

        closest_idx = min(range(len(levels)), key=lambda t: abs(levels[t] - frequency))
        y_values = [0] * (len(levels) - closest_idx)
        overfit_area = np.trapz(phase_rand_curve[closest_idx:], levels[closest_idx:])
        plt.fill_between(levels[closest_idx:], phase_rand_curve[closest_idx:],
                         y_values,
                         where=(phase_rand_curve[closest_idx:] >= 0),
                         color='blue', alpha=0.5)
        zone_image = f'{self.va_dir}/feature_zone.png'
        plt.savefig(zone_image)
        plt.close()

        return feature_area, overfit_area
    @staticmethod
    def area_difference(curve_one, curve_two, levels):
        """
            Calculates the area difference between two curves share common x axis (levels)
        :param curve_one: array 0f curve one
        :param curve_two: array 0f curve two
        :param levels: array 0f levels
        """
        area = np.trapz(curve_one - curve_two, levels)

        return area

    def feature_zone(self, data_curves):
        """
            By given two curves get the feature zone based on the intersections
            return quantified area of that zone
            :param data_curves: dictionary contains all curves
        """

        zones = {}
        result = {}
        all_curves = None
        if 'curves' in data_curves.keys():
            all_curves = data_curves['curves']

        levels = None
        corrected_curve = None
        phase_rand_curve = None
        masked_curve = None
        unmasked_curve = None
        if all_curves:
            if 'level' in all_curves.keys():
                levels = np.array(all_curves['level'])
            if 'fsc_corrected' in all_curves.keys():
                corrected_curve = np.array(all_curves['fsc_corrected'])
            if 'phaserandmization' in all_curves.keys():
                phase_rand_curve = np.array(all_curves['phaserandmization'])
            if 'phaserandomization' in all_curves.keys():
                phase_rand_curve = np.array(all_curves['phaserandomization'])
            if 'fsc_masked' in all_curves.keys():
                masked_curve = np.array(all_curves['fsc_masked'])
            if 'fsc' in all_curves.keys():
                unmasked_curve = np.array(all_curves['fsc'])

        if isinstance(levels, np.ndarray) and isinstance(corrected_curve, np.ndarray) and isinstance(phase_rand_curve, np.ndarray):
            xs, ys = interpolated_intercepts_general(levels, corrected_curve, phase_rand_curve)
            intersections = remove_duplicate_intersections(xs, ys)
            nlevels, ncorrected_curve, phase_rand_curve = self.intersections_into_curve(intersections, levels, corrected_curve, phase_rand_curve)
            feature_area, overfit_area = self.plot_feature_zone(intersections, nlevels, ncorrected_curve, phase_rand_curve)
            masking_area = self.area_difference(masked_curve, corrected_curve, levels)
            corrected_unmasked_difference = sum(corrected_curve - unmasked_curve)
            corrected_masked_difference = sum(masked_curve - corrected_curve)
            if corrected_unmasked_difference == 0:
                print('Corrected is the same as the unmasked.')
                masking_area_ratio = 999
            else:
                masking_area_ratio = abs(corrected_masked_difference / corrected_unmasked_difference)
            zones['feature_zone'] = keep_three_significant_digits(feature_area)
            zones['overfit_zone'] = keep_three_significant_digits(overfit_area)
            zones['masking_area'] = keep_three_significant_digits(masking_area)
            zones['masking_area_ratio'] = keep_three_significant_digits(masking_area_ratio)
            zones['feature_zone_ratio'] = keep_three_significant_digits(zones['feature_zone'] / (zones['feature_zone'] + zones['overfit_zone']))
            result['feature_zones'] = zones

        return result











import subprocess
import sys
import os
from distutils.spawn import find_executable


def create_folder(output_path):
    """
        create emringer output folder inside va directory

    :return: model related path of strudel folder
    """

    fullname = '{}'.format(output_path)

    if not os.path.isdir(fullname):
        os.mkdir(fullname, mode=0o777)
    else:
        print('{} is exist'.format(fullname))


def run_emringer(full_modelpath, full_mappath, output_path):
    """

        Emringer score

    :return:
    """

    errlist = []
    try:
        assert find_executable('phenix.emringer') is not None
        emringerpath = find_executable('phenix.emringer')
        create_folder(output_path)
        try:
            subprocess.check_call(emringerpath + ' ' + full_modelpath + ' ' + full_mappath, cwd=output_path, shell=True)
        except subprocess.CalledProcessError as suberr:
            err = 'EMRinger calculation error: {}.'.format(suberr)
            errlist.append(err)
            sys.stderr.write(err + '\n')
    except AssertionError:
        sys.stderr.write('emringer executable is not there.\n')

    return errlist


def emringerpkl_json(self):
    """

        Load EMRiinger output pickle file and write out to a JSON file

    :return:
    """

    pass



import subprocess
from va.utils.misc import *
from distutils.spawn import find_executable
from va.metrics.contour_level_predicator import *
import mrcfile

def create_relion_folders(root, mapname, result_type):
    """
    Create <mapname>_relion folder in the va directory and subfolders containing mask, fsc and local_resolution
    """
    relion_dir = f'{root}{mapname}_relion'
    mask_dir = f'{relion_dir}/mask'
    fsc_dir = f'{relion_dir}/fsc'
    local_resolution_dir = f'{relion_dir}/local_resolution'

    if create_directory(relion_dir) and create_directory(mask_dir):
        print('Relion folder and mask folder created.')
        if result_type == 'fsc':
            if create_directory(fsc_dir):
                print('FSC folder created.')
                return fsc_dir, mask_dir
            else:
                print('FSC folder was not created. Please check.')
                return None, None

        if result_type == 'locres':
            if create_directory(local_resolution_dir):
                print('FSC folder created.')
                return local_resolution_dir, mask_dir
            else:
                print('FSC folder was not created. Please check.')
                return None, None
        print('Check the input folder type either fsc or locres.')
        return None, None
    else:
        print('Relion folder and mask folder was not created. Please check')
        return None, None


def check_mrc(input_map):
    """
    Check if the input map is in MRC format. If not, create a symbolic link in the same folder ends with mrc
    """

    suffix = '.mrc'
    if not input_map.endswith('.mrc'):
        last_dot_index = input_map.rfind('.')
        mrc_input_map = input_map[:last_dot_index] + suffix
        if create_symbolic_link(input_map, mrc_input_map):
            return mrc_input_map
    else:
        print('MRC file format.')
        return None

def relion_mask(mapone, maptwo, out_dir, relion_mask_executable, mapname=None):

    original_input_mrc = check_mrc(mapone)
    angpix = get_voxel_size(mapone)
    input_mrc = mrcfile.open(original_input_mrc)
    d = input_mrc.data
    max_val_thirty = f'{calc_level_dev(d)[0] * 0.7}'

    dilatepx, softpx = calculate_pixels(angpix)
    mask_loose = os.path.join(out_dir, f'{mapname}_mask.mrc')

    if angpix and max_val_thirty and dilatepx and softpx and input_mrc:

        subprocess.run(f'{relion_mask_executable} --i {original_input_mrc} --o {mask_loose} --ini_threshold {max_val_thirty} '
                       f'--extend_inimask {dilatepx} --width_soft_edge {softpx} --lowpass 12 ', shell=True)

        return mask_loose
    else:
        print(f'!!! Check the cmd: {relion_mask_executable} --i {original_input_mrc} --o {mask_loose} --ini_threshold {max_val_thirty} '
                       f'--extend_inimask {dilatepx} --width_soft_edge {softpx} --lowpass 12')
        return None


def relion_fsc(mapone, maptwo, out_dir, relion_executable, mapname=None):
    """
    Relion FSC calculation running
    """

    mapone_mrc = check_mrc(mapone)
    maptwo_mrc = check_mrc(maptwo)
    mask_file = f'{os.path.dirname(out_dir)}/mask/{mapname}_mask.mrc'

    if mapone_mrc and maptwo_mrc and mask_file:

        subprocess.run(f"{relion_executable} --i {mapone_mrc} --i2 {maptwo_mrc} --o {out_dir}/fsc"
                           f" --mask {mask_file} --auto_bfac", shell=True)

        return True
    else:
        print(f'!!! Check the cmd: f"{relion_executable} --i {mapone_mrc} --i2 {maptwo_mrc} --o {out_dir}/fsc"'
              f' --mask {mask_file} --auto_bfac", shell=True)"')
        return None

def relion_fsc_calculation(mapone, maptwo, root, mapname=None):
    """
    Calculates FSc using Relion

    return: data dictionary of fsc
    """

    result_type = 'fsc'
    if not mapname:
        mapname = f'{os.path.basename(mapone)}_{os.path.basename(maptwo)}'
    fsc_dir, mask_dir = create_relion_folders(root, mapname, result_type)
    relion_executable_name = 'relion_postprocess'
    relion_mask_executable_name = 'relion_mask_create'

    try:
        assert find_executable(relion_executable_name) and find_executable(relion_mask_executable_name) is not None
        relion_executable = find_executable(relion_executable_name)
        relion_mask_executable = find_executable(relion_mask_executable_name)
        if fsc_dir and mask_dir and relion_mask_executable:
            relion_mask(mapone, maptwo, mask_dir, relion_mask_executable, mapname)
            if relion_fsc(mapone, maptwo, fsc_dir, relion_executable, mapname):
                return f'{fsc_dir}/fsc.star'
            else:
                return None

    except Exception as e:
        print('No relion or fsc calculation was wrong.')
        return None

def get_voxel_size(input_map):

    try:
        with mrcfile.open(input_map, permissive=True) as mrc:
            voxel_size = mrc.voxel_size

            return float(voxel_size['x'])
    except Exception as e:
        print('No voxel size of the map.')
        return None


# def get_max_value(input_map):
#     """
#     Get max value from the input map
#     """
#
#     try:
#         with mrcfile.open(input_map, permissive=True) as mrc:
#             dmax = mrc.header.dmax
#
#             return float(dmax)
#     except Exception as e:
#         print('No max value of the input map.')
#         return None


def calculate_pixels(angpix):
    """
    Calculate hard and soft pixel for mask map
    """

    if angpix != 0:
        dilatepx = 10 / angpix
        softpx = 5 / angpix
        return dilatepx, softpx
    else:
        print('No hard and soft radius for mask as voxel value is 0.')
        return None, None


# def read_fscs(star):
#     # Read the block, note data_fsc becomes fsc
#     fsc_block = cif.read_file(str(star)).find_block("fsc")
#
#     # Read the loop, note loop_ excluded and only search by suffix
#     resolution = list(fsc_block.find_loop("_rlnResolution"))
#     resolutionAng = list(fsc_block.find_loop("_rlnAngstromResolution"))
#     fsc = list(fsc_block.find_loop("_rlnFourierShellCorrelationCorrected"))
#     fsc_unmasked = list(fsc_block.find_loop("_rlnFourierShellCorrelationUnmaskedMaps"))
#     fsc_randomised = list(fsc_block.find_loop("_rlnCorrectedFourierShellCorrelationPhaseRandomizedMaskedMaps"))
#
#     # Create a DataFrame for the fsc values
#     data = {
#         'resolution': pd.Series(resolution, dtype=float),
#         'resolutionAng': pd.Series(resolutionAng, dtype=float),
#         'fsc_corrected': pd.Series(fsc, dtype=float),
#         'fsc_unmasked': pd.Series(fsc_unmasked, dtype=float),
#         'fsc_randomised': pd.Series(fsc_randomised, dtype=float)
#     }
#     df = pd.DataFrame(data)
#
#     return df

# def assess_fscs(star, df):
#     print('Assessing post processing star file')
#     ## Postprocessing information
#     # Read the block, note data_
#     block = cif.read_file(star).find_block("general")
#
#     # Read star file pairs, excluded and only search by suffix
#     randomise = block.find_pair("_rlnRandomiseFrom")[1]
#
#     bfac = block.find_pair("_rlnBfactorUsedForSharpening")[1]
#
#     # Look up the 'resolution' value corresponding to the 'randomise' value in df
#     resolution_randomise = df[df['resolutionAng'] == float(randomise)]['resolution'].values
#
#     # Check if the resolution corresponding to randomise value is found in df
#     if resolution_randomise:
#         resolution_randomise = resolution_randomise[0]
#     else:
#         resolution_randomise = 0
#
#     ## Do some sanity checks on the FSC curves
#     # Report
#     print()
#     print('### Phase randomisation assessment:')
#     print()
#     print(f"The resolution (Ang) value where phases were randomised is: {randomise}")
#     print(f"The corresponding resolution value is: {resolution_randomise}")
#     print()
#
#     # Filter the DataFrame for FSC_randomised values below 0.143
#     filtered_df = df[df['fsc_randomised'] < 0.143]
#     # Check if there are rows in the filtered DataFrame
#     if not filtered_df.empty:
#         randomisation_drop = 'normal behaviour'
#         # Retrieve the resolution value when FSC drops below 0.143
#         resolutionAng_rdm_below_0_143 = filtered_df.iloc[0]['resolutionAng']
#         resolution_rdm_below_0_143 = filtered_df.iloc[0]['resolution']
#         print(f"The resolution (Ang) value where FSC_randomised drops below 0.143 is: {resolutionAng_rdm_below_0_143}")
#         print(f"The corresponding resolution value is: {resolution_rdm_below_0_143}")
#         print()
#
#         # Is the rate of drop from phase randomisation rapid?
#         differenceAng = float(randomise) - float(resolutionAng_rdm_below_0_143)
#         difference = float(resolution_rdm_below_0_143) - float(resolution_randomise)
#         print(f"The FSC_randomisation drops from 1-0.143 in x (Ang): {differenceAng}")
#         print(f"The FSC_randomisation drops from 1-0.143 in x (1/Ang): {difference}")
#         print('Low values are better, indicating a rapid drop in the phase randomised curve after phase randomisation')
#         print()
#         # This is an example check and needs validating to establish what the correct normal and abnormal behaviour is
#         if difference > 0.05:
#             randomisation_rate = 'abnormal behaviour'
#         else:
#             randomisation_rate = 'normal behaviour'
#
#     else:
#         randomisation_rate = 'N.D.'
#         randomisation_drop = 'abnormal behaviour'
#
#     # Phase randomisation reporting
#     print('1: FSC_randomised drops below 0.143: '+randomisation_drop)
#     print('2: FSC_randomised drop rate from 1-0.143: '+randomisation_rate)
#
#     print()
#     print('### Resolution assessment:')
#     print()
#
#     measure_resolution(df, 0.143, 'fsc_randomised', 'FSC_randomised')
#     measure_resolution(df, 0.143, 'fsc_unmasked', 'FSC_unmasked')
#     measure_resolution(df, 0.143, 'fsc_corrected', 'FSC_masked_corrected')
#
#     print()
#
#     # Declare naughty global variables
#     assess_fscs.bfac = bfac

# def measure_resolution(df, threshold, column, label):
#     # Filter the DataFrame for FSC values below 0.143
#     filtered_df = df[df[column] < threshold]
#     # Retrieve the resolution value when FSC drops below 0.143
#     resolution_below_0_143 = filtered_df.iloc[0]['resolutionAng']
#
#     print(f"The resolution (Ang) value where FSC drops below "+str(threshold)+" for "+str(label)+": "+str(resolution_below_0_143))


# def generate_mask(mapname, mapone, maptwo, out_dir):
#
#     angpix = get_voxel_size(mapone)
#     dilatepx, softpx = calculate_pixels(angpix)
#
#     output_directory = os.path.join(out_dir, f'{mapname}_relion_phrand')
#     mask_loose = os.path.join(output_directory, f'{mapname}_mask_loose.mrc')
#     os.makedirs(output_directory, exist_ok=True)
#
#     if not os.path.isfile(mask_loose):
#         output = subprocess.run(f"relion_image_handler --i {mapone} --stats", shell=True, capture_output=True,
#                                 text=True)
#         mapvalmax = output.stdout.split('=')[5].split(';')[0]
#
#         mapval25 = float(mapvalmax) * 0.30
#         mapval25_formatted = f"{mapval25:.3f}"
#
#         print(f"The 30% of mapvalmax is: {mapval25_formatted}")
#
#         subprocess.run(
#             f"relion_mask_create --i {mapone} --o {mask_loose} --ini_threshold {mapval25_formatted} --extend_inimask {dilatepx} --width_soft_edge {softpx}",
#             shell=True)
#
#         return mask_loose
#     else:
#         print("mask_loose.mrc already exists. Skipping mask creation.")
#         return None


# def generate_star(mapname, mapone, maptwo, out_dir):
#
#     mask_loose = generate_mask(mapname, mapone, out_dir)
#     postprocess_star_file = os.path.join(out_dir, f'{mapname}_postprocess.star')
#     if not os.path.isfile(postprocess_star_file) and os.path.isfile(mask_loose):
#         subprocess.run(f"relion_postprocess --i {mapone} --i2 {maptwo} --o {out_dir}/{mapname}_relion_phrand --mask {mask_loose} --auto_bfac",
#             shell=True)
#     else:
#         print("postprocess.star already exists. Skipping post-processing.")
#
#
#     if os.path.isfile(postprocess_star_file):
#         return postprocess_star_file
#     else:
#         return None

# def create_mrclink(mapone, maptwo):
#
#     cmapone = mapone[:-2] + 'rc'
#     cmaptwo = maptwo[:-2] + 'rc'
#     if not os.path.isfile(cmapone):
#         create_symbolic_link(mapone, cmapone)
#     if not os.path.isfile(cmaptwo):
#         create_symbolic_link(maptwo, cmaptwo)
#
#     return cmapone, cmaptwo
#

# def plot_fsc(df, plt_name):
#     # Set the style to whitegrid
#     plt.style.use('seaborn-whitegrid')
#
#     # Create a figure and axis objects
#     fig, ax = plt.subplots(figsize=(10, 6))
#
#     # Plot scatter plots
#     ax.scatter(df["resolution"], df["fsc_corrected"], label="FSC_masked_corrected", color='blue')
#     ax.scatter(df["resolution"], df["fsc_unmasked"], label="FSC_unmasked", color='grey')
#     ax.scatter(df["resolution"], df["fsc_randomised"], label="FSC Randomised", color='orange')
#
#     # Plot line plots
#     ax.plot(df["resolution"], df["fsc_corrected"], color='blue')
#     ax.plot(df["resolution"], df["fsc_unmasked"], color='grey')
#     ax.plot(df["resolution"], df["fsc_randomised"], color='orange')
#
#     # Add a horizontal line at FSC value 0.143
#     ax.axhline(y=0.143, color='red', linestyle='--', label='FSC = 0.143')
#
#     # Set title and labels
#     ax.set_title('FSC Corrected')
#     ax.set_xlabel('Resolution (1/Ang)')
#     ax.set_ylabel('FSC Values')
#
#     # Add legend
#     ax.legend()
#
#     # Save the plot
#     plt.savefig(plt_name)



# def plot_fsc(df, plt_name):
#     ## Plotting the scatter plot with lines
#     sns.set(style="whitegrid")
#
#     plt.figure(figsize=(10, 6))
#
#     sns.scatterplot(data=df, x="resolution", y="fsc_corrected", label="FSC_masked_corrected", color='blue')
#     sns.scatterplot(data=df, x="resolution", y="fsc_unmasked", label="FSC_unmasked", color='grey')
#     sns.scatterplot(data=df, x="resolution", y="fsc_randomised", label="FSC Randomised", color='orange')
#
#     sns.lineplot(data=df, x="resolution", y="fsc_corrected", color='blue')
#     sns.lineplot(data=df, x="resolution", y="fsc_unmasked", color='grey')
#     sns.lineplot(data=df, x="resolution", y="fsc_randomised", color='orange')
#
#     # Add a horizontal line at FSC value 0.143
#     plt.axhline(y=0.143, color='red', linestyle='--', label='FSC = 0.143')
#
#     plt.title('FSC Corrected')
#     plt.xlabel('Resolution (1/Ang)')
#     plt.ylabel('FSC Values')
#
#     plt.legend()
#     plt.savefig(plt_name)
# def post_phrand(mapname, mapone, maptwo, out_dir):
#
#     onemrc, twomrc = create_mrclink(mapone, maptwo)
#     ## Define the relion postprocess.star file for assessment
#     star = generate_star(mapname, onemrc, twomrc, out_dir)
#
#     ## Read FSC values into dataframe
#     if star:
#         df = read_fscs(star)
#
#         ## Do FSC curve assessments
#         assess_fscs(star, df)
#
#     # Check if the postprocess.star file already exists
#     if not star:
#         subprocess.run(f"relion_postprocess --i {input_file1} --i2 {input_file2} --o {output_directory}/postprocess --mask {mask_loose} --adhoc_bfac {assess_fscs.bfac} --locres", shell=True)
#     else:
#         print("postprocess.star already exists. Skipping post-processing.")
#
#     ## plot FSC curves
#     plot_name = f'{out_dir}/{mapname}_relion_fsc.png'
#     plot_fsc(df, plot_name)


import os
from os import path
import sys
from va.version import __version__

sys.path.append(path.dirname(path.dirname(path.abspath(__file__))))
# sys.path.append(path.dirname(path.abspath(__file__)))
BASE_DIR = os.path.dirname(os.path.realpath(__file__))

name = 'va'


from va.utils.misc import floatohex, scale_values, scale_value, float_to_hex
from va.utils.MapProcessor import MapProcessor
from va.utils.ChimeraxViews import ChimeraxViews


def local_resolution_json(map, inmodels, radius, value_range):
    """
        Interpolate density value of one atom, if indices are on the same plane use nearest method
        otherwise use linear

    :param map: a string of full map path or a mrcfile object
    :param inmodels: Structure instance from biopython package mmcif parser
    :param radius: a number to define the radius that take into account around per atom
    :param value_range: tuple of local resolution value range used to colour the map
    :return: a dictionary with all residue local resolution values
    """

    map = MapProcessor(map)
    models = []
    if isinstance(inmodels, list):
        models = inmodels
    else:
        models.append(inmodels)
    result = {}
    modelcount = 0
    for model in models:
        modelcount += 1
        modelname = model.filename.split('/')[-1]
        atomcount = 0
        chainaiscore = {}
        data_result = {}
        allkeys = []
        allvalues = []
        colors = []
        chainai_atomsno = {}
        # use for generate mask for the whole model or the voxels involved
        # tmp_allindices = set()
        # tmp_allcoords = []
        for chain in model.get_chains():
            chain_name = chain.id
            chain_residue_count = 0
            chain_resolution = 0.
            for residue in chain.get_residues():
                residue_atom_count = 0
                chain_residue_count += 1
                residue_name = residue.resname
                residue_no = residue.id[1]
                nearresidue_voxels = set()
                for atom in residue.get_atoms():
                    if atom.name.startswith('H') or atom.get_parent().resname == 'HOH':
                        continue
                    atomcount += 1
                    residue_atom_count += 1
                    onecoor = atom.coord
                    around_indices = map.get_close_voxels_indices(onecoor, radius)
                    nearresidue_voxels.update(around_indices)
                    # use for generate mask for the whole model or the voxels involved
                    # tmp_allcoords.append(onecoor)
                    # tmp_allindices.update(around_indices)
                if residue_atom_count == 0:
                    continue
                # residue inclusion section
                keystr = f'{chain_name}:{residue_no} {residue_name}'
                allkeys.append(keystr)
                cur_residue_resolution = map.residue_average_resolution(list(nearresidue_voxels))
                value = float('%.4f' % cur_residue_resolution)
                residue_color = floatohex([value])[0]
                chain_resolution += value
                allvalues.append(value)
                colors.append(residue_color)
            # chain inclusion section
            if chain_name in chainai_atomsno.keys():
                chain_resolution += chainai_atomsno[chain_name]['value']
                chain_residue_count += chainai_atomsno[chain_name]['residuesinchain']
            # For cases where one water molecule has a single but different chain id
            if chain_residue_count == 0:
                continue
            chainai_atomsno[chain_name] = {'value': chain_resolution, 'residuesinchain': chain_residue_count}

        # produce a mask for entire model
        # map.generate_mask(tmp_allcoords, radius)

        for chainname, chain_scores in chainai_atomsno.items():
            chain_ai = float('%.3f' % round((float(chain_scores['value']) / chain_scores['residuesinchain']), 4))
            aicolor = floatohex([chain_ai])[0]
            chainaiscore[chainname] = {'value': chain_ai, 'color': aicolor, 'numberOfResidues': chain_scores['residuesinchain']}
        # For using new_min and new_max which match to 0 and 1 for model-local-resolution colouring
        # map_min = map.map.data.min()
        # map_max = map.map.data.max()
        map_min, map_max = value_range
        residues_min = min(allvalues)
        residues_max = max(allvalues)
        real_min = map_min if map_min < residues_min else residues_min
        real_max = map_max if map_max > residues_max else residues_max
        new_min = scale_value(residues_min, real_min, real_max, 0, 1)
        new_max = scale_value(residues_max, real_min, real_max, 0, 1)

        data_result['residue'] = allkeys
        data_result['localResolution'] = allvalues
        data_result['color'] = [float_to_hex(i) for i in scale_values(allvalues, new_min, new_max)]
        # data_result['color'] = [float_to_hex(i) for i in scale_values(allvalues, 0, 1)]
        data_result['chainResolution'] = chainaiscore
        result[str(modelcount-1)] = {'name': modelname, 'data': data_result}

    return result

def model_local_resolution_views(input_json, map_name, data_type, chimerax_bin_dir):
    """
        Generate local resolution views based on the input json file

    :param input_json: a string of full path to input json file
    :param map_name: a string of map name
    :param data_type: a string of type
    """

    viewer = ChimeraxViews(chimerax_bin_dir, input_json)
    root_data = viewer.get_root_data(data_type)
    viewer.get_model_views(map_name, root_data, data_type)

def map_local_resolution_views(map_name, data_type, chimerax_bin_dir, work_dir, input_json=None):
    """
        Generate local resolution view based on local resolution map
    """
    if input_json:
        viewer = ChimeraxViews(chimerax_bin_dir, input_json)
        root_data = viewer.get_root_data('residue_local_resolution')
        viewer.get_map_views(map_name, root_data, data_type)
    else:
        viewer = ChimeraxViews(chimerax_bin_dir, None, work_dir)
        # viewer.get_views(map_name,None, data_type)
        viewer.get_map_views(map_name,None, data_type)




import subprocess
import sys
import os
from distutils.spawn import find_executable
import timeit
import time


def create_folder(output_path):
    """
        create strudel output folder inside va directory

    :return: model related path of strudel folder
    """

    fullname = '{}'.format(output_path)

    if not os.path.isdir(fullname):
        os.mkdir(fullname, mode=0o777)
    else:
        print('{} is exist'.format(fullname))


def run_monitored_job(command, work_dir):
    """
    Submits a job to the queue system and waits until it finishes
    :param command: Bsub command
    :param work_dir: directory where the job will be run
    """
    command = command.split()
    # Submit job to the queue ang get the job id
    out = str(subprocess.check_output(command, cwd=work_dir))
    job_id = out.split('<')[1].split('>')[0]
    while True:
        # Check the job status every 20 seconds
        try:
            out1 = str(subprocess.check_output(["bjobs", "-a", "-noheader", job_id]))
            status = out1.split()[2]
        except subprocess.CalledProcessError:
            continue
        if status == 'DONE' or status == 'EXIT':
            break
        elif 'not found' in out1:
            break
        else:
            time.sleep(20)


def run_strudel(full_modelpath, full_mappath, motif_libpath, output_path, queue, platform=None):
    """
        full_modelpath: full path of the model with name
        full_mappath: full path of the map with name
        motif_libpath: full path of the motif lib for its resolution
        output_path: output directory
    :return:
    """

    start = timeit.default_timer()
    create_folder(output_path)
    num_processors = int(os.cpu_count() / 2)
    bsub_bin = find_executable('bsub')
    strudel_cmd = 'strudel_mapMotifValidation.py -p {} -m {} -l {} -o {} -np {}'.format(full_modelpath,
                                                                                            full_mappath, motif_libpath,
                                                                                            output_path, num_processors)
    if platform == 'emdb' and bsub_bin:
        strudel_cmd = ('sbatch --mem=26G --time 48:00:00 -p {} -J strudel --ntasks=8 --mail-user=zhe@ebi.ac.uk '
                       '--mail-type=END --mail-type=FAIL -e {}/strudel_stderr.txt -o {}/strudel_stdout.txt '
                       '--wrap="strudel_mapMotifValidation.py -p {} -m {} -l {} -o {} -np 8 '
                       '-log {}/strudel.log -r -rs"').format(
            queue,
            output_path,
            output_path,
            full_modelpath,
            full_mappath,
            motif_libpath,
            output_path, output_path)
    errlist = []
    print(strudel_cmd)
    try:
        # run_monitored_job(strudel_cmd, output_path)
        # subprocess.check_call(strudel_cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, shell=True,
        #                                 cwd=output_path)
        process = subprocess.Popen(strudel_cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, shell=True,
                                   cwd=output_path)
        # out = subprocess.check_output(strudel_cmd.split(), stderr=subprocess.STDOUT, shell=True, cwd=output_path)
        # print(out)

        output = process.communicate('n\n')[0]
        errstrudelscore = 'error'
        if sys.version_info[0] >= 3:
            for item in output.decode('utf-8').split('\n'):
                print(item)
                if errstrudelscore in item.lower():
                    errline = item.strip()
                    errlist.append(errline)
                    assert errstrudelscore not in output.decode('utf-8'), errline

        else:
            for item in output.split('\n'):
                print(item)
                if errstrudelscore in item.lower():
                    errline = item.strip()
                    errlist.append(errline)
                    assert errstrudelscore not in output.decode('utf-8'), errline

        end = timeit.default_timer()
        print('Strudel time: %s' % (end - start))
        print('------------------------------------')
    except:
        end = timeit.default_timer()
        err = 'Strudel error: {}'.format(sys.exc_info()[1])
        errlist.append(err)
        sys.stderr.write(err + '\n')
        print('Strudel time: %s' % (end - start))
        print('------------------------------------')


    # If needed here will do the post processing of strudel score
    # posts_trudlescore()

    return errlist


def strudel_tojson(self):
    """
        Process strudel score related files

    :return:
    """

    pass



import os
import codecs
import json
import numpy as np
from PIL import Image


def change_filename(input_filename, output_filename):
    try:
        os.rename(input_filename, output_filename)
        print(f"File name changed from '{input_filename}' to '{output_filename}'")
    except FileNotFoundError:
        print(f"Error: File '{input_filename}' not found.")
    except FileExistsError:
        print(f"Error: File '{output_filename}' already exists.")

def create_directory(dir):
    """
    Create a directory if it doesn't exist and return the dir otherwise None

    :return: str or None
    """

    try:
        os.makedirs(dir, exist_ok=True)

        return dir
    except Exception as e:
        print(f'Create folder error: {e}')

        return None


def create_symbolic_link(file_path, link_name):
    """
    Create symbolic link to a file

    :return: str or None
    """

    file_dir = os.path.dirname(file_path)
    link_path = os.path.join(file_dir, link_name)

    if os.path.isfile(link_path):
        print(f"Symbolic link '{link_name}' already exists")
        return link_path
    else:
        try:
            os.symlink(file_path, link_path)
            print(f"Symbolic link '{link_name}' created successfully")
            return link_path
        except Exception as e:
            print(f"Error creating symbolic link: {e}")

            return None


def interpolated_intercept(x, y1, y2):
    """

        Find the intercept of two curves, given by the same x data

    """

    def intercept(point1, point2, point3, point4):
        """

            Find the intersection between two lines
            the first line is defined by the line between point1 and point2
            the second line is defined by the line between point3 and point4
            each point is an (x,y) tuple.

            So, for example, you can find the intersection between
            intercept((0,0), (1,1), (0,1), (1,0)) = (0.5, 0.5)

            :return: Intercept, in (x,y) format

        """

        def line(p1, p2):
            A = (p1[1] - p2[1])
            B = (p2[0] - p1[0])
            C = (p1[0] * p2[1] - p2[0] * p1[1])

            return A, B, -C

        def intersection(L1, L2):
            D = L1[0] * L2[1] - L1[1] * L2[0]
            Dx = L1[2] * L2[1] - L1[1] * L2[2]
            Dy = L1[0] * L2[2] - L1[2] * L2[0]

            x = Dx / D
            y = Dy / D

            return x, y

        L1 = line([point1[0], point1[1]], [point2[0], point2[1]])
        L2 = line([point3[0], point3[1]], [point4[0], point4[1]])

        R = intersection(L1, L2)

        return R

    if not isinstance(x, np.ndarray):
        x = np.asarray(x)
    if not isinstance(y1, np.ndarray):
        y1 = np.asarray(y1)
    if not isinstance(y2, np.ndarray):
        y2 = np.asarray(y2)
    idx = np.argwhere(np.diff(np.sign(y1 - y2)) != 0)
    # Remove the first point usually (0, 1) to avoid all curves starting from (0, 1) which pick as one intersection
    if idx.size != 0:
        if idx[0][0] == 0:
            idx = np.delete(idx, 0, 0)
        xc, yc = intercept((x[idx], y1[idx]), ((x[idx + 1], y1[idx + 1])), ((x[idx], y2[idx])),
                           ((x[idx + 1], y2[idx + 1])))
        return xc, yc
    else:
        nullarr = np.empty(shape=(0, 0))
        return nullarr, nullarr


def interpolated_intercepts_general(x, y1, y2):
    """Find the intercepts of two curves, given by the same x data"""

    def intercept(point1, point2, point3, point4):
        """Find the intersection between two lines.

        The first line is defined by the line between point1 and point2.
        The second line is defined by the line between point3 and point4.
        Each point is an (x, y) tuple.

        Returns: the intercept, in (x, y) format.
        """

        def line(p1, p2):
            A = (p1[1] - p2[1])
            B = (p2[0] - p1[0])
            C = (p1[0] * p2[1] - p2[0] * p1[1])
            return A, B, -C

        def intersection(L1, L2):
            D = L1[0] * L2[1] - L1[1] * L2[0]
            Dx = L1[2] * L2[1] - L1[1] * L2[2]
            Dy = L1[0] * L2[2] - L1[2] * L2[0]
            x = Dx / D
            y = Dy / D
            return x, y

        L1 = line([point1[0], point1[1]], [point2[0], point2[1]])
        L2 = line([point3[0], point3[1]], [point4[0], point4[1]])

        return intersection(L1, L2)

    idxs = np.argwhere(np.diff(np.sign(y1 - y2)) != 0)


    xcs = []
    ycs = []

    for idx in idxs:
        xc, yc = intercept((x[idx], y1[idx]), (x[idx + 1], y1[idx + 1]), (x[idx], y2[idx]), (x[idx + 1], y2[idx + 1]))

        txc = xc[0].tolist()
        tyc = yc[0].tolist()

        xcs.append(txc)
        ycs.append(tyc)

    return xcs, ycs

def remove_duplicate_intersections(xs, ys):
    """
        Given two lists of xs and ys pair them and remove the duplicates pairs
        return unique pairs
        :param xs: list of x values
        :param ys: list of y values
    """

    rxs = [round(val, 5) for val in xs]
    rys = [round(val, 5) for val in ys]
    # round xs and ys to 4-5 decimal potins and then do the following
    pairs = list(zip(rxs, rys))
    unique_pairs = set(pairs)
    unique_pairs_list = sorted(list(unique_pairs), key=lambda x: x[0])

    return unique_pairs_list


def out_json(data_dict, out_file_path):
    """
    Given the data_dict and out_file_path, write the data to a json file
    """

    try:
        with codecs.open(out_file_path, 'w', encoding='utf-8') as outfile:
            json.dump(data_dict, outfile)
        print("Data successfully written to", out_file_path)
    except Exception as e:
        print("Error:", e)


def first_non_zero_index(number_str):
    """
    Find the index of the first non-zero digit in a string representation of a number.
    """

    for i, digit in enumerate(number_str):
        if digit != '0':
            return i
    return -1


def keep_three_significant_digits(number, significant_digits=3):
    """
    Keep three significant digits without using scientific notation.
    """

    number_float = float(number)

    # Check if the number is an integer
    if number_float.is_integer():
        return int(number_float)  # Return the integer part as a string

    if number_float > 1:
        return round(number_float, significant_digits)
    else:
        # Convert the float to a string with the desired precision
        number_str = '{:.10f}'.format(number_float).rstrip('0')  # Avoid trailing zeros

        # Split the number into integer and fractional parts
        integer_part, decimal_part = number_str.split('.')
        non_zero_index = first_non_zero_index(decimal_part)
        if non_zero_index == -1:
            return round(number_float, significant_digits)
        else:
            final_decimal_part = decimal_part[:non_zero_index + significant_digits]
            final = f'{integer_part}.{final_decimal_part}'

            # Return the number with the desired number of significant digits
            return float(final)

def floatohex(numlist):
    """
        Produce hex color between red and green
    :param numlist: A list of RGB values
    :return: A list of hex value between R and G with B = 0
    """

    numlist = [-1 if i < 0 else i for i in numlist]
    rgbs = [[122, int(num * 255), int(num * 255)] if num >= 0 else [255, 0, 255] for num in numlist]
    resultlist = ['#%02X%02X%02X' % (rgb[0], rgb[1], rgb[2]) for rgb in rgbs]

    return resultlist


def float_to_hex(value, platelet=None):
    """
        Convert float to hex color if platelet is give using the platelet

    :param value: float between 0 and 1
    :param platelet: list of tuple with rgb values or default as None use rainbow

    """
    if not platelet:
        # default as rainbow
        colors = [
            # rainbow
            # (128, 0, 128),  # Violet
            # (75, 0, 130),  # Indigo
            # (0, 0, 255),  # Blue
            # (0, 128, 128),  # Blue-Green
            # (0, 255, 0),  # Green
            # (128, 255, 0),   # Yellow-Green
            # (255, 255, 0),  # Yellow
            # (255, 165, 0),  # Orange
            # (255, 69, 0),   # Orange-Red
            # (255, 0, 0)  # Red

            # bwr
            (0, 0, 255),  # Blue
            (255, 255, 255),  # White
            (255, 0, 0)  # Red
        ]
    else:
        colors = platelet

    index = int(value * (len(colors) - 1))
    if index >= len(colors) - 1:
        index = len(colors) - 2
    start_color = colors[index]
    end_color = colors[index + 1]
    fraction = (value - index / (len(colors) - 1)) * (len(colors) - 1)
    interpolated_color = [
        int(start + fraction * (end - start)) for start, end in zip(start_color, end_color)
    ]
    hex_color_code = '#{:02x}{:02x}{:02x}'.format(*interpolated_color)

    return hex_color_code


def check_same_keys(dict1, dict2):
    """
        check if two dictionaries have the same keys at all levels
    dict1: A dictionary
    dict2: A dictionary
    return True if all keys are same, else False
    """

    if not isinstance(dict1, dict) or not isinstance(dict2, dict):
        return False

    keys1 = set(dict1.keys())
    keys2 = set(dict2.keys())

    if keys1 != keys2:
        return False

    for key in keys1:
        if isinstance(dict1[key], dict) and isinstance(dict2[key], dict):
            if not check_same_keys(dict1[key], dict2[key]):
                return False
        else:
            if not isinstance(dict1[key], dict) and not isinstance(dict2[key], dict):
                continue
            else:
                return False

    return True


def scale_values(values, a, b):
    """
    scales values between a and b
    :param values: list of values
    :param a: value of lower bound
    :param b: value of upper bound
    """
    min_val = min(values)
    max_val = max(values)

    # Scale each value to the specified range [a, b]
    # scaled_values = [((val - min_val) / (max_val - min_val)) * (b - a) + a for val in values]
    scaled_values = [scale_value(val, min_val, max_val, a, b) for val in values]

    return scaled_values

def scale_value(value, a, b, c, d):
    # Scale value from range (a, b) to range (c, d)
    scaled_value = (value - a) * (d - c) / (b - a) + c
    return scaled_value

def inverse_scale_value(scaled_value, a, b, c, d):
    # Scale value from range (c, d) to range (a, b)
    inverse_scaled_value = (scaled_value - c) * (b - a) / (d - c) + a
    return inverse_scaled_value


def scale_image(image, scale):
    """
    """
    npimg = Image.open(image)
    im = npimg.resize(scale)
    image_name = os.path.basename(image)
    image_parts = image_name.split('_')
    first_part = '_'.join(image_parts[:-1])
    out_image_name = f'{first_part}_scaled_{image_parts[-1]}'
    dir_name = os.path.dirname(image)
    scaled_image = f'{dir_name}/{out_image_name}'
    im.save(scaled_image)

    return out_image_name


def pad_array(input_array, output_shape, pad_value=0):
    """
        Pad the input array to the specified output shape with the given pad value.

    :param input_array: the input array to be padded
    :param output_shape: the desired shape of the output array
    :param pad_value: the value used for padding
    """
    input_shape = input_array.shape
    pad_width = [(max(d_out - d_in, 0), max(d_in - d_out, 0)) for d_in, d_out in zip(input_shape, output_shape)]
    padded_array = np.pad(input_array, pad_width, mode='constant', constant_values=pad_value)

    return padded_array



import os
os.environ['KMP_DUPLICATE_LIB_OK'] = 'True'
import torch
import torch.nn as nn
import numpy as np
torch.manual_seed(42)
import warnings
import mrcfile

warnings.filterwarnings("ignore", message="CUDA initialization: The NVIDIA driver on your system is too old.*")
warnings.filterwarnings("ignore", category=RuntimeWarning)
warnings.filterwarnings("ignore", message="User provided device_type of 'cuda', but CUDA is not available. Disabling")


# Convolutional layers adapted from https://www.nature.com/articles/s41598-022-19212-6
class CNNModel(nn.Module):
    def __init__(self):
        super(CNNModel, self).__init__()
        self.conv = nn.Sequential(
            nn.Conv3d(1, 32, 7, 1, padding_mode='replicate'),
            nn.BatchNorm3d(32),
            nn.ReLU(),
            nn.Conv3d(32, 64, 5, 1, padding_mode='replicate'),
            nn.BatchNorm3d(64),
            nn.ReLU(),
            nn.MaxPool3d(2),
            nn.Conv3d(64, 128, 5, 1, padding_mode='replicate'),
            nn.BatchNorm3d(128),
            nn.ReLU(),
            nn.MaxPool3d(2),
            nn.Conv3d(128, 256, 3, 1, padding_mode='replicate'),
            nn.BatchNorm3d(256),
            nn.ReLU(),
            nn.MaxPool3d(2),
            nn.Conv3d(256, 384, 1, 1, padding_mode='replicate'),
            nn.ReLU(),
            nn.AvgPool3d(3, stride=2),
            nn.Dropout(0.3))

        self.mlp = nn.Sequential(nn.Linear(in_features=384, out_features=64),
                                 nn.ReLU(),
                                 nn.Linear(in_features=64, out_features=1))

    def forward(self, x):
        x = x.view(-1, 1, 64, 64, 64)
        x = self.conv(x)
        x = x.view(x.size(0), -1)
        x = self.mlp(x)
        return x


def check_boxsize(vol):
    return vol.shape[0]


def load_data(vol_d_norm):
    vol_array = np.empty(shape=(1, 64, 64, 64), dtype=np.float32)
    vol_array[:] = vol_d_norm
    data_generator = torch.utils.data.DataLoader(vol_array, batch_size=1, shuffle=False)
    return data_generator


def downsample(vol):
    boxsize = check_boxsize(vol)
    vol_ft = np.fft.fftshift(np.fft.fftn(np.fft.fftshift(vol)))
    start = boxsize // 2 - 64 // 2
    stop = boxsize // 2 + 64 // 2
    vol_ft_d = vol_ft[start:stop, start:stop, start:stop]
    vol_d = np.fft.ifftshift(np.fft.ifftn(np.fft.ifftshift(vol_ft_d)))
    vol_d = vol_d.real
    vol_d *= ((64 / boxsize) ** 3)
    vol_d = vol_d.astype(np.float32)
    return vol_d


def upsample(
        vol):  # through interpolation of maps in Fourier space; based on https://www.sciencedirect.com/science/article/pii/S001046551400335X
    boxsize = check_boxsize(vol)
    new_box_size = 64
    vol_ft = np.fft.fftshift(np.fft.fftn(np.fft.fftshift(vol)))
    padded_vol_ft = np.zeros((new_box_size, new_box_size, new_box_size), dtype=np.float32)
    padded_vol_ft[:vol.shape[0], :vol.shape[1], :vol.shape[2]] = vol_ft
    interpolated_density_map = np.fft.ifftshift(np.fft.ifftn(np.fft.ifftshift((padded_vol_ft))))
    interpolated_density_map = interpolated_density_map.real
    interpolated_density_map *= ((64 / boxsize) ** 3)
    return interpolated_density_map


def normalize(vol_data):
    d_upper = np.percentile(vol_data, 99.999)
    d_lower = np.percentile(vol_data, 0.001)
    vol_data = np.where(vol_data > d_upper, d_upper, vol_data)
    vol_data = np.where(vol_data < d_lower, d_lower, vol_data)
    vol_d_norm = np.where(vol_data <= 0, 0, vol_data)
    vol_d_min = np.min(vol_data[vol_data > 0])
    vol_d_max = np.max(vol_d_norm)
    vol_d_norm = (vol_d_norm - vol_d_min) / (vol_d_max - vol_d_min)
    vol_d_norm = 2 * vol_d_norm - 1
    return vol_d_norm, vol_d_min, vol_d_max


device = torch.device("cuda") if torch.cuda.is_available() else torch.device('cpu')
model = CNNModel()
dir = os.path.dirname(os.path.abspath(__file__))
parent_dir = os.path.dirname(dir)
weights_dir = os.path.join(parent_dir, 'utils')
checkpoint = torch.load(os.path.join(weights_dir, 'cl_weights.pth'), map_location=device)
model.load_state_dict(checkpoint['model_state_dict'])


def model_pred(data_generator):
    model.eval()
    pred_list = []
    with torch.no_grad():
        with torch.cuda.amp.autocast():
            for (idx, x) in enumerate(data_generator):
                x = x.to(device)
                prediction = model(x)
                prediction = (prediction.squeeze(dim=1))
                pred_list.append(prediction)
    return pred_list


def calc_level_dev(vol_data):
    boxsize = check_boxsize(vol_data)

    if boxsize > 64:
        _, vol_min, vol_max = normalize(vol_data)
        vol_d = downsample(vol_data)
        vol_d_norm, _, _ = normalize(vol_d)

    elif boxsize < 64:
        _, vol_min, vol_max = normalize(vol_data)
        vol_d = upsample(vol_data)
        vol_d_norm, _, _ = normalize(vol_d)

    else:
        vol_d = vol_data
        vol_d_norm, vol_d_min, vol_d_max = normalize(vol_d)

    data_generator = load_data(vol_d_norm)
    pred_list = model_pred(data_generator)
    all_preds = np.concatenate(pred_list, axis=0)
    int_preds = (all_preds + 1) / 2

    if boxsize != 64:
        norm_preds = (int_preds * (vol_max - vol_min)) + vol_min
    else:
        norm_preds = (int_preds * (vol_d_max - vol_d_min)) + vol_d_min

    return norm_preds









import json
import os
import re
import sys
import glob
import subprocess
import itertools
from distutils.spawn import find_executable
from va.utils.misc import scale_image, out_json
from va.metrics.contour_level_predicator import *
from va.utils.MapProcessor import *

class ChimeraxViews:

    def __init__(self, chimerax_bin_dir, input_json=None, va_dir=None):
        self.input_json = input_json
        self.va_dir = os.path.dirname(self.input_json) if input_json else va_dir
        if self.input_json:
            with open(input_json, 'r') as f:
                self.json_data = json.load(f)
        else:
            self.json_data = None
            print('There is no json file data.')

        self.chimerax = chimerax_bin_dir


    def get_root_data(self, data_type):
        """
            Get the root json data based on the input json file and the data type

        :param data_type: a string of full path name of the json file
        """

        root_data = None
        if data_type in self.json_data.keys():
            root_data = self.json_data[data_type]
            try:
                del root_data['err']
            except:
                print('There is/are %s model(s).' % len(root_data))
        else:
            print(f'{data_type} json result is')

        return root_data

    def json_to_attributes(self, data_type):
        """
            Generate the attribute file for ChimeraX to colour models
        :param data_type: a string of data type
        todo: function need to be adjusted according to general data_type

        """
        data_type = 'residue_local_resolution'
        root_data = self.get_root_data(data_type)
        result = []
        for i in range(0, len(root_data)):
            first_model_name = root_data[str(i)]['name']
            first_model_data = root_data[str(i)]['data']
            residues = first_model_data['residue']
            local_resolutions = first_model_data['localResolution']
            colors = first_model_data['color']
            n = len(list((set(colors))))
            local_resolution_colors = list(zip(local_resolutions, colors))
            sorted_local_resolution_colors = sorted(local_resolution_colors)
            number_colours = int(0.1 * n)
            step = (len(sorted_local_resolution_colors) - 1) / (number_colours - 1)
            selected_elements = [sorted_local_resolution_colors[int(round(step * i))] for i in range(number_colours)]
            f = ''
            for i in selected_elements:
                f += f'{i[0]},{i[1]}:'
            palette = f[:-1]

            file_name = f'{self.va_dir}/{first_model_name}_{data_type.replace("_", "")}.defattr'
            cur_file = open(file_name, 'w')
            cur_file.write(f'attribute: {data_type.replace("_", "")}\n')
            cur_file.write(f'recipient: residues\n')
            for residue, local_resolution in zip(residues, local_resolutions):
                residue_no = (residue.split(' ')[0]).split(':')[1]
                line = f'\t:{residue_no}\t{round(local_resolution, 1)}\n'
                cur_file.write(line)
            result.append((file_name, palette))

        return result

    def write_residue_attr_cxc(self, map_name, model_name, data_type, min_median_max, palette, attr_file):
        """
            Write a ChimeraX cxc file for generating surfaces with model cases
        :param map_name: a string of input file name
        :param model_name: a string of input model name
        :param data_type: a string of surface type, e.g., residue_local_resolution
        :param min_median_max: list of 3 tuples contains min value, colour, max value, colour and median value, colour
        :param palette: a string of palette used in ChimeraX
        :param attr_file: a string to attributes file produced by json_to_attributes
        """

        if data_type:
            cur_type = data_type.replace('_', '')
            chimerax_file_name = f'{map_name}_{model_name}_{cur_type}_chimerax.cxc'
        else:
            cur_type = ''
            chimerax_file_name = f'{map_name}_{model_name}_chimerax.cxc'

        surface_file_name = '{}/{}_{}'.format(self.va_dir, map_name, model_name)
        model = f'{self.va_dir}/{model_name}'
        list_items = list(itertools.islice(min_median_max.items(), 3))
        min_value, min_color = list_items[0]
        median_value, median_color = list_items[1]
        max_value, max_color = list_items[2]
        with open(f'{self.va_dir}/{chimerax_file_name}', 'w') as fp:
            fp.write(f'open {model} format mmcif\n')
            fp.write(f'open {attr_file}\n')
            fp.write('show selAtoms ribbons\n')
            fp.write('hide selAtoms\n')
            fp.write(
                # f'color byattribute {data_type.replace("_", "")} palette {palette}\n'
                # 'key showTool true\n'
                f'color byattribute {data_type.replace("_", "")} palette bluered key True\n'
                'key size 0.25000,0.02000\n'
                'key pos 0.72000,0.05000\n'
                # f'key {min_color}:{round(min_value, 1)} {median_color}:{round(median_value, 1)} {max_color}:{round(max_value, 1)}\n'
                'set bgColor white\n'
                'lighting soft\n'
                'view cofr True\n'
                f'save {str(surface_file_name)}_z{cur_type}.jpeg supersample 3 width 1200 height 1200\n'
                'turn x -90\n'
                'turn y -90\n'
                'view cofr True\n'
                f'save {str(surface_file_name)}_y{cur_type}.jpeg supersample 3 width 1200 height 1200\n'
                'view orient\n'
                'turn x 90\n'
                'turn z 90\n'
                'view cofr True\n'
                f'save {str(surface_file_name)}_x{cur_type}.jpeg supersample 3 width 1200 height 1200\n'
                'close all\n'
                'exit'
            )

            return chimerax_file_name

    def write_residue_cxc(self, colors, residues, map_name, model_name, data_type, min_median_max):
        """
            Write a ChimeraX cxc file for generating surfaces with model cases
        :param colors: list of colors
        :param residues: list of residues e.g., A:321 THR
        :param map_name: a string of input file name
        :param model_name: a string of input model name
        :param data_type: a string of surface type, e.g., residue_local_resolution
        """

        if data_type:
            cur_type = data_type.replace('_', '')
            chimerax_file_name = f'{map_name}_{model_name}_{cur_type}_chimerax.cxc'
        else:
            cur_type = ''
            chimerax_file_name = f'{map_name}_{model_name}_chimerax.cxc'

        surface_file_name = '{}/{}_{}'.format(self.va_dir, map_name, model_name)
        model = f'{self.va_dir}/{model_name}'
        list_items = list(itertools.islice(min_median_max.items(), 3))
        min_value, min_color = list_items[0]
        median_value, median_color = list_items[1]
        max_value, max_color = list_items[2]
        with open(f'{self.va_dir}/{chimerax_file_name}', 'w') as fp:
            fp.write(f'open {model} format mmcif\n')
            fp.write('show selAtoms ribbons\n')
            fp.write('hide selAtoms\n')

            for (color, residue) in zip(colors, residues):
                chain, restmp = residue.split(':')
                # Not sure if all the letters should be replaced
                # res = re.sub("\D", "", restmp)
                res = re.findall(r'-?\d+', restmp)[0]
                fp.write(
                    f'color /{chain}:{res} {color}\n'
                )
            fp.write(
                'key showTool true\n' 
                'key size 0.25000,0.02000\n'
                'key pos 0.72000,0.05000\n'
                # f'key blue-white-red :{round(min, 2)} :{round((min+max)/2, 2)} :{round(max, 2)}\n'
                f'key {min_color}:{round(min_value, 1)} {median_color}:{round(median_value, 1)} {max_color}:{round(max_value, 1)}\n'
                'set bgColor white\n'
                'lighting soft\n'
                'view cofr True\n'
                f'save {str(surface_file_name)}_z{cur_type}.jpeg supersample 3 width 1200 height 1200\n'
                'turn x -90\n'
                'turn y -90\n'
                'view cofr True\n'
                f'save {str(surface_file_name)}_y{cur_type}.jpeg supersample 3 width 1200 height 1200\n'
                'view orient\n'
                'turn x 90\n'
                'turn z 90\n'
                'view cofr True\n'
                f'save {str(surface_file_name)}_x{cur_type}.jpeg supersample 3 width 1200 height 1200\n'
                'close all\n'
                'exit'
            )

            return chimerax_file_name

    def write_maps_cxc(self, map_name, map_two, palette, data_type):
        """
            Write a ChimeraX cxc file for generating surfaces with model cases
        :param map_name: a string of input file name
        :param map_two: a string of input model name
        :param data_type: a string of surface type, e.g., residue_local_resolution
        """

        if data_type:
            cur_type = data_type.replace('_', '')
            chimerax_file_name = f'{map_name}_{cur_type}_chimerax.cxc'
        else:
            cur_type = ''
            chimerax_file_name = f'{map_name}_chimerax.cxc'

        map_fullname = f'{self.va_dir}/{map_name[:-4]}.map'
        # mask_fullname = f'{self.va_dir}/{map_name}_relion/mask/{map_name}_mask.mrc'

        surface_file_name = '{}/{}'.format(self.va_dir, map_name)
        palette_list = palette.split(':')
        _, min_color = palette_list[0].split(',')
        _, max_color = palette_list[-1].split(',')
        middle_index = (len(palette_list) - 1) // 2
        _, median_color = palette_list[middle_index].split(',')

        with open(f'{self.va_dir}/{chimerax_file_name}', 'w') as fp:
            fp.write(f'open {map_fullname} format ccp4\n')
            fp.write(f'open {map_two} format ccp4\n')
            if data_type == 'map_local_resolution':
                a = mrcfile.open(map_fullname)
                d = a.data
                raw_contour = f'{calc_level_dev(d)[0]}'
                min_value, max_value = MapProcessor.map_minmax(map_fullname, map_two)
                median_value = keep_three_significant_digits((min_value + max_value) * 0.5)
                fp.write(f'volume #1 step 1 level {raw_contour}\n')
                fp.write(f'color sample #1 map #2 palette bluered key true\n')
                # Here is use the real range for colouring the scale bar
                # fp.write(f'key showTool true\n')
                # fp.write(f'color sample #1 map #2 palette {palette}\n')
                # fp.write(f'key {min_color}:{round(float(min_value), 1)} {median_color}:{round(float(median_value), 1)} {max_color}:{round(float(max_value), 1)}\n')
                fp.write('key size 0.25000,0.02000\n')
                fp.write('key pos 0.72000,0.05000\n')
                fp.write('hide #!2 models\n')
            if data_type == 'map_mask':
                # prepare for map and mask overlay views here
                pass

            fp.write(
                'set bgColor white\n'
                'lighting soft\n'
                'view cofr True\n'
                f'save {str(surface_file_name)}_z{cur_type}.jpeg supersample 3 width 1200 height 1200\n'
                'turn x -90\n'
                'turn y -90\n'
                'view cofr True\n'
                f'save {str(surface_file_name)}_y{cur_type}.jpeg supersample 3 width 1200 height 1200\n'
                'view orient\n'
                'turn x 90\n'
                'turn z 90\n'
                'view cofr True\n'
                f'save {str(surface_file_name)}_x{cur_type}.jpeg supersample 3 width 1200 height 1200\n'
                'close all\n'
                'exit'
                # First line: show on z axis plane 220 with step 1 at contour level 0.01 with style surface
                # 2nd line: as 1st line is one plane, 2nd line turn it back to full surface with contour level 3.5
                #           at step 1 so it can continue the same 2 lines to get other axis planes
                # 'volume #1 planes z,220 step 1 level 0.01 style surface'
                # 'volume #1 style surface region all level 3.5 step 1'
            )

            return chimerax_file_name

    def run_chimerax(self, chimerax_file_name):
        """
            Run ChimeraX to produce the surface views

        :param chimerax_file_name: a string of ChimeraX cxc file
        """
        errs = []
        chimerax = self.chimerax
        model_name = chimerax_file_name.split('_')[1]
        bin_display = os.getenv('DISPLAY')
        try:
            if not bin_display:
                subprocess.check_call(f'{chimerax} --offscreen --nogui {self.va_dir}/{chimerax_file_name}',
                                      cwd=self.va_dir, shell=True)
                print('Colored models were produced.')
            else:
                subprocess.check_call(f'{chimerax}  {self.va_dir}/{chimerax_file_name}', cwd=self.va_dir, shell=True)
                print('Colored models were produced.')

            return None
        except subprocess.CalledProcessError as e:
            err = 'Saving model {} fit surface view error: {}.'.format(model_name, e)
            errs.append(err)
            sys.stderr.write(err + '\n')

            return errs

    def rescale_view(self, map_name, model_name=None, data_type=None):
        """
            Scale views and produce corresponding dictionary

        :param map_name: a string of input map name
        :param model_name: a string of input model name
        :param data_type: a string of view type
        """

        original = {}
        scaled = {}
        result = {}
        used_data_type = data_type.replace('_', '')
        for i in 'xyz':
            if model_name is None:
                image_name = f'{map_name}_{i}{used_data_type}.jpeg'
            else:
                image_name = f'{map_name}_{model_name}_{i}{used_data_type}.jpeg'
            full_image_path = f'{self.va_dir}/{image_name}'
            if os.path.isfile(full_image_path):
                scaled_image_name = scale_image(full_image_path, (300, 300))
                original[i] = image_name
                scaled[i] = scaled_image_name
        result['original'] = original
        result['scaled'] = scaled

        return result

    def get_model_views(self, map_name, root_data=None, data_type=None):
        """
            Based on the information produce views and save to json file

        :param root_data: root data from input json file
        :param map_name: a string of input map name
        :param data_type: a string of view type
        """

        if root_data:
            # Model coloured based on local resolution
            num_model = len(root_data)
            for i in range(num_model):
                output_json = {}
                json_dict = {}
                cur_model = root_data[str(i)]
                keylist = list(cur_model.keys())
                colors = None
                residues = None
                values = None
                model_name = None
                for key in keylist:
                    if key != 'name':
                        colors = cur_model[key]['color']
                        residues = cur_model[key]['residue']
                        values = cur_model[key]['localResolution']
                    else:
                        model_name = cur_model[key]
                min_middle_max = {}
                min_value = min(values)
                max_value = max(values)
                real_middle = (min_value + max_value) / 2
                difference = [abs(value - real_middle) for value in values]
                middle_index = difference.index(min(difference))
                middle_value = values[middle_index]
                min_value, middle_value, max_value = (min(values), middle_value, max(values))
                min_index = values.index(min_value)
                max_index = values.index(max_value)
                min_colour = colors[min_index]
                middle_colour = colors[middle_index]
                max_colour = colors[max_index]
                min_middle_max[min_value] = min_colour
                min_middle_max[middle_value] = middle_colour
                min_middle_max[max_value] = max_colour
                result = self.json_to_attributes(data_type)
                # chimerax_file_name = self.write_residue_cxc(colors, residues, map_name, model_name, data_type, min_middle_max)
                # Use attributes file for colouring
                chimerax_file_name = self.write_residue_attr_cxc(map_name, model_name, data_type, min_middle_max, result[0][1], result[0][0])
                out = self.run_chimerax(chimerax_file_name)
                surfaces_dict = self.rescale_view(map_name, model_name, data_type)
                json_dict[model_name] = surfaces_dict
                output_json[f'{data_type}_views'] = json_dict

                output_json_file = f"{map_name}_{model_name}_{data_type.replace('_', '')}.json"
                output_json_fullpath = f'{self.va_dir}/{output_json_file}'
                out_json(output_json, output_json_fullpath)

    def get_map_views(self, map_name, root_data=None, data_type=None):
        """
            Get the views from ChimeraX
        :param map_name: string of full map name (primary map)
        :param root_data: json root data to that specific data_type
        :param data_type: string of data type e.g., 'qscore'
        """

        all_models_min_value = 999
        all_models_max_value = -1
        if root_data:
            # Model coloured based on local resolution
            num_model = len(root_data)
            for i in range(num_model):
                cur_model = root_data[str(i)]
                keylist = list(cur_model.keys())
                values = None
                for key in keylist:
                    if key != 'name':
                        values = cur_model[key]['localResolution']
                    else:
                        model_name = cur_model[key]
                min_value = min(values)
                max_value = max(values)
                if min_value < all_models_min_value:
                    all_models_min_value = min_value
                if max_value > all_models_max_value:
                    all_models_max_value = max_value
        output_json = {}
        raw_map = f'{self.va_dir}/{map_name[:-4] + "_rawmap.map"}'
        mask_map = f'{self.va_dir}/{map_name}_relion/mask/{map_name}_mask.mrc'
        local_resolution_map_glob = glob.glob(f'{self.va_dir}/{map_name}_relion/*_locres.mrc')
        local_resolution_map = local_resolution_map_glob[0] if len(local_resolution_map_glob) > 0 else None
        map_processor = MapProcessor()
        binarized_mask_map = map_processor.binarized_mask(mask_map, map_name)
        masked_raw_map = map_processor.mask_map(raw_map, binarized_mask_map)
        map_min, map_max = map_processor.map_minmax(masked_raw_map, local_resolution_map)
        real_min = map_min if map_min < all_models_min_value else all_models_min_value
        real_max = map_max if map_max > all_models_max_value else all_models_max_value
        new_min = scale_value(map_min, real_min, real_max, 0, 1)
        new_max = scale_value(map_max, real_min, real_max, 0, 1)
        num_elements = 50
        palette_colours = self.generate_palette(new_min, new_max, map_min, map_max, num_elements)
        chimerax_file_name = self.write_maps_cxc(os.path.basename(masked_raw_map), local_resolution_map, palette_colours, data_type)
        out = self.run_chimerax(chimerax_file_name)
        surfaces_dict = self.rescale_view(os.path.basename(masked_raw_map), None, data_type)
        output_json[f'{data_type}_views'] = surfaces_dict

        output_json_file = f"{map_name}_{data_type.replace('_', '')}.json"
        output_json_fullpath = f'{self.va_dir}/{output_json_file}'
        out_json(output_json, output_json_fullpath)

    @staticmethod
    def generate_palette(new_min, new_max, map_min, map_max, length):
        """
            generate a string of colours palette used in ChimeraX
        :param new_min: map_min match to 0 and 1
        :param new_max: map_max match to 0 and 1
        :param map_min: min value of map data
        :param map_max: max value of map data
        """

        real_palette_values = [keep_three_significant_digits(i, 1) for i in np.linspace(map_min, map_max, length).tolist()]
        palette_values = np.linspace(new_min, new_max, length).tolist()
        color_palette_values = [keep_three_significant_digits(i) for i in palette_values]
        palette_colours = [float_to_hex(i) for i in color_palette_values]
        f = ''
        for i, j in zip(real_palette_values, palette_colours):
            f += f'{i},{j}:'
        palette = f[:-1]

        return palette



from emda.ext.cc import cc_overall_realsp
from emda.ext.cc import cc_overall_fouriersp


def run_realmmcc(mapa, mapb, mask=None):
    """
        Calculate the real space map-modelmap correlation

    :param mapa: primary map
    :param mapb: map from the model
    :param mask: mask map if available
    :return: cross-correlation value
    """

    realcc = cc_overall_realsp(mapa, mapb, mask)

    return realcc





import subprocess
import sys
from distutils.spawn import find_executable
import numpy as np
import mrcfile
from PIL import Image, ImageSequence
import glob
import json
import codecs
from va.utils.misc import *

import matplotlib.pyplot as plt


def run_locres(mapone, maptwo, output_path, resmappy_path=None):
    errlist = []
    try:
        bindisplay = os.getenv('DISPLAY')
        mapone_name = os.path.basename(mapone)
        maptwo_name = os.path.basename(maptwo)
        locres_map = f'{output_path}/{mapone_name}_{maptwo_name}'
        one_linkname = mapone_name[:-2] + 'rc'
        two_linkname = maptwo_name[:-2] + 'rc'
        run_mapone_name = create_symbolic_link(mapone, one_linkname)
        run_maptwo_name = create_symbolic_link(maptwo, two_linkname)
        if bindisplay:
            assert find_executable('relion_postprocess') is not None
            respath = find_executable('relion_postprocess')
            resmap_cmd = '{} --i {} --i2 {} --o {} --locres --locres_sampling 30'.format( respath, run_mapone_name, run_maptwo_name, locres_map)
            print(resmap_cmd)
        else:
            if resmappy_path == None:
                raise Exception('No ResMap.py path provided.')
            else:
                resmap_cmd = '{} --i {} --i2 {} --o {} --locres --locres_sampling 30'.format( resmappy_path, run_mapone_name, run_maptwo_name, locres_map)
                print(resmap_cmd)
        create_directory(output_path)
        try:
            process = subprocess.Popen(resmap_cmd.split(), stdout=subprocess.PIPE, stderr=subprocess.STDOUT,
                                       cwd=output_path)
            output = process.communicate('n\n')[0]
            errqscore = 'error'
            for item in output.decode('utf-8').split('\n') if sys.version_info[0] >= 3 else output.split('\n'):
                print(item)
                if errqscore in item.lower():
                    errline = item.strip()
                    errlist.append(errline)
                    assert errqscore not in output.decode('utf-8'), errline

            # if sys.version_info[0] >= 3:
            #     for item in output.decode('utf-8').split('\n'):
            #         # item = cline.decode('utf-8').strip()
            #         print(item)
            #         if errqscore in item.lower():
            #             errline = item.strip()
            #             errlist.append(errline)
            #             assert errqscore not in output.decode('utf-8'), errline
            #
            # else:
            #     for item in output.split('\n'):
            #         print(item)
            #         if errqscore in item.lower():
            #             errline = item.strip()
            #             errlist.append(errline)
            #             assert errqscore not in output.decode('utf-8'), errline
        except subprocess.CalledProcessError as suberr:
            err = 'Local resolution from Relion calculation error: {}.'.format(suberr)
            errlist.append(err)
            sys.stderr.write(err + '\n')
    except AssertionError as exerr:
        err = 'Relion executable is not there.'
        errlist.append(err)
        sys.stderr.write('Relion executable is not there: {}\n'.format(exerr))

    return errlist


def locres_filecheck(mapone, maptwo, output_path):
    """
        Check if all output from ResMap are ready
    :return: True(all file exist) or False (not all file exist
    """

    output_path = output_path
    mapname_one = os.path.basename(mapone)
    mapname_two = os.path.basename(maptwo)
    starfile = '{}/{}_{}_locres_fscs.star'.format(output_path, mapname_one, mapname_two)
    locres_name = '{}/{}_{}_locres.mrc'.format(output_path, mapname_one, mapname_two)
    locres_filtered = '{}/{}_{}_locres_filtered.mrc'.format(output_path, mapname_one, mapname_two)
    check = os.path.isfile(starfile) and os.path.isfile(locres_name) and os.path.isfile(locres_filtered)

    return check if check else False


def run_resmap(mapone, maptwo, output_path, resmappy_path=None):
    errlist = []
    try:
        bindisplay = os.getenv('DISPLAY')
        if bindisplay:
            assert find_executable('ResMap') is not None
            respath = find_executable('ResMap')
            resmap_cmd = '{} {} {} --noguiSplit --doBenchMarking'.format(respath, mapone, maptwo)
            print(resmap_cmd)
        else:
            if resmappy_path == None:
                raise Exception('No ResMap.py path provided.')
            else:
                resmap_cmd = '{} {} {} {} --noguiSplit --doBenchMarking'.format('python', resmappy_path, mapone, maptwo)
                print(resmap_cmd)
        create_folder(output_path)
        try:
            process = subprocess.Popen(resmap_cmd.split(), stdout=subprocess.PIPE, stderr=subprocess.STDOUT,
                                       cwd=output_path)
            output = process.communicate('n\n')[0]
            errqscore = 'error'
            if sys.version_info[0] >= 3:
                for item in output.decode('utf-8').split('\n'):
                    # item = cline.decode('utf-8').strip()
                    print(item)
                    if errqscore in item.lower():
                        errline = item.strip()
                        errlist.append(errline)
                        assert errqscore not in output.decode('utf-8'), errline

            else:
                for item in output.split('\n'):
                    print(item)
                    if errqscore in item.lower():
                        errline = item.strip()
                        errlist.append(errline)
                        assert errqscore not in output.decode('utf-8'), errline
        except subprocess.CalledProcessError as suberr:
            err = 'Local resolution from ResMap calculation error: {}.'.format(suberr)
            errlist.append(err)
            sys.stderr.write(err + '\n')
    except AssertionError as exerr:
        err = 'ResMap executable is not there.'
        errlist.append(err)
        sys.stderr.write('ResMap executable is not there: {}\n'.format(exerr))

    return errlist

def resmap_filecheck(mapone, output_path):
    """
        Check if all output from ResMap are ready
    :return: True(all file exist) or False (not all file exist
    """

    output_path = output_path
    mapname = os.path.basename(mapone)
    mapfile = '{}{}_ori.map'.format(output_path, os.path.splitext(mapname)[0])
    resmapname = '{}{}_ori_resmap.map'.format(output_path, os.path.splitext(mapname)[0])
    resmap_chimera = '{}{}_ori_resmap_chimera.cmd'.format(output_path, os.path.splitext(mapname)[0])
    check = os.path.isfile(mapfile) and os.path.isfile(resmapname) and os.path.isfile(resmap_chimera)

    return check if check else False


def resmap_chimerax(mapone, output_path):
    """
        Generate chimerax cmd for ResMap results
    :return:
    """

    mapname = os.path.basename(mapone)
    output_chimerax_file = '{}{}_chimerax.cxc'.format(output_path, os.path.basename(mapone))
    orgmap = '{}{}_ori.map'.format(output_path, os.path.splitext(mapname)[0])
    resmap = '{}{}_ori_resmap.map'.format(output_path, os.path.splitext(mapname)[0])
    header = mrcfile.open(mapone, mode='r', header_only=True)
    voxsizes = header.voxel_size.tolist()
    if all(element == voxsizes[0] for element in voxsizes):
        voxsize = voxsizes[0]
        mmin = round((2.2 * voxsize) / 0.1) * 0.1  # round to the nearest 0.1
        mmax = round((4.0 * voxsize) / 0.5) * 0.5  # round to the nearest 0.5
    else:
        voxsizemin = voxsizes.min()
        voxsizemax = voxsizes.max()
        mmin = round((2.2 * voxsizemin) / 0.1) * 0.1  # round to the nearest 0.1
        mmax = round((4.0 * voxsizemax) / 0.5) * 0.5  # round to the nearest 0.5
        print('Voxel sizes are not the same!!!')

    n = header.header.nx
    colors_chimerax = ['blue', 'cyan', 'green', 'yellow', 'orange', 'red']
    values_chimerax = np.linspace(mmin, mmax, len(colors_chimerax))
    colorstr = ''
    for i in range(len(colors_chimerax)):
        colorstr += '%.2f' % values_chimerax[i] + ',' + colors_chimerax[i] + ':'
    colorstr += '%.2f' % (values_chimerax[i] + 0.01) + ',' + 'gray'
    with open(output_chimerax_file, 'w') as fp:
        fp.write("windowsize 800 800\n")
        fp.write("set bg_color white\n")
        fp.write("open " + orgmap + " format ccp4" + '\n')
        fp.write("open " + resmap + " format ccp4" + '\n')
        fp.write("volume #2 hide\n")
        fp.write("color sample #1 map #2 palette " + colorstr + '\n')
        fp.write("volume #1 planes z,0 step 1 level -1 style surface\n")
        # non-perspective mode is not working with Chimerax 1.6 on headless machine
        # fp.write("camera ortho\n")
        fp.write("zoom 1.3\n")
        fp.write("view cofr True\n")
        # fp.write("camera ortho\n")
        fp.write('2dlabels text "0" xpos 0.8 ypos 0.1\n')
        # fp.write("vop gaussian #1 sDev 5 model #3")
        # fp.write("volume #3 level 0.02 step 1 color rgba(0%, 80%, 40%, 0.5)")
        # fp.write("volume #3 level 0.02 step 1 color 60,0,80,50")
        # fp.write("zoom 0.5")
        fp.write("movie record size 800,800\n")
        # fp.write('perframe "volume #1 plane z,$1" range 198,0')
        fp.write('perframe "volume #1 plane z,$1 ; 2dlabel #3.1 text $1" range 0,{},2\n'.format(n))
        fp.write("wait {}\n".format(str(round(n/2) + 2)))
        fp.write("movie encode {}_zplanes_noloop.png quality high\n\n\n\n".format(resmap))

        # X
        fp.write("close session\n")
        fp.write("set bg_color white\n")
        fp.write("open " + orgmap + " format ccp4" + '\n')
        fp.write("open " + resmap + " format ccp4" + '\n')
        fp.write("volume #2 hide\n")
        fp.write("color sample #1 map #2 palette " + colorstr + '\n')
        fp.write("turn -x 90\n")
        fp.write("turn -y 90\n")
        fp.write("volume #1 planes x,0 step 1 level -1 style surface\n")
        # fp.write("camera ortho\n")
        fp.write("zoom 1.3\n")
        fp.write("view cofr True\n")
        # fp.write("view orient\n")
        # non-perspective mode is not working with Chimerax 1.6 on headless machine
        # fp.write("camera ortho\n")
        fp.write('2dlabels text "0" xpos 0.8 ypos 0.1\n')
        # fp.write("vop gaussian #1 sDev 5 model #3")
        # fp.write("volume #3 level 0.02 step 1 color rgba(0%, 80%, 40%, 0.5)")
        # fp.write("volume #3 level 0.02 step 1 color 60,0,80,50")
        # fp.write("zoom 0.5")
        fp.write("movie record size 800,800\n")
        # fp.write('perframe "volume #1 plane z,$1" range 198,0')
        fp.write('perframe "volume #1 plane x,$1 ; 2dlabel #3.1 text $1" range 0,{},2\n'.format(n))
        fp.write("wait {}\n".format(str(round(n/2) + 2)))
        fp.write("movie encode {}_xplanes_noloop.png quality high\n".format(resmap))

        # Y
        fp.write("close session\n")
        fp.write("set bg_color white\n")
        fp.write("open " + orgmap + " format ccp4" + '\n')
        fp.write("open " + resmap + " format ccp4" + '\n')
        fp.write("volume #2 hide\n")
        fp.write("color sample #1 map #2 palette " + colorstr + '\n')
        fp.write("turn x 90\n")
        fp.write("turn z 90\n")
        fp.write("volume #1 planes y,0 step 1 level -1 style surface\n")
        # non-perspective mode is not working with Chimerax 1.6 on headless machine
        # fp.write("camera ortho\n")
        fp.write("zoom 1.3\n")
        fp.write("view cofr True\n")
        # fp.write("view orient\n")
        # fp.write("camera ortho\n")
        fp.write('2dlabels text "0" xpos 0.8 ypos 0.1\n')
        # fp.write("vop gaussian #1 sDev 5 model #3")
        # fp.write("volume #3 level 0.02 step 1 color rgba(0%, 80%, 40%, 0.5)")
        # fp.write("volume #3 level 0.02 step 1 color 60,0,80,50")
        # fp.write("zoom 0.5")
        fp.write("movie record size 800,800\n")
        # fp.write('perframe "volume #1 plane z,$1" range 198,0')
        fp.write('perframe "volume #1 plane y,$1 ; 2dlabel #3.1 text $1" range 0,{},2\n'.format(n))
        fp.write("wait {}\n".format(str(round(n/2) + 2)))
        fp.write("movie encode {}_yplanes_noloop.png quality high\n".format(resmap))
        fp.write('close all\n')
        fp.write('exit')
        fp.close()

        return output_chimerax_file

def run_resmap_chimerax(bindisplay, locCHIMERA, cxcfile):
    """

    :return:
    """
    errlist = []
    try:
        if not bindisplay:
            subprocess.check_call(locCHIMERA + " --offscreen --nogui " + cxcfile, cwd=os.path.dirname(cxcfile),
                                  shell=True)
            print('Animated PNG for ResMap result has been produced.')
        else:
            subprocess.check_call(locCHIMERA + " " + cxcfile, cwd=os.path.dirname(cxcfile), shell=True)
            print('Animated PNG for ResMap result has been produced.')
    except subprocess.CalledProcessError as suberr:
        err = 'Saving ResMap local resolution animated png error: {}.'.format(suberr)
        errlist.append(err)
        sys.stderr.write(err + '\n')

    return errlist

def save_imagestojson(resmapname, output_filename):
    """

    :return:
    """
    basename = os.path.basename(resmapname)
    x_local = '{}_xplanes_noloop.png'.format(resmapname)
    y_local = '{}_yplanes_noloop.png'.format(resmapname)
    z_local = '{}_zplanes_noloop.png'.format(resmapname)
    x_localo = '{}_xplanes.png'.format(resmapname)
    y_localo = '{}_yplanes.png'.format(resmapname)
    z_localo = '{}_zplanes.png'.format(resmapname)
    make_apng_looping(x_local, x_localo)
    make_apng_looping(y_local, y_localo)
    make_apng_looping(z_local, z_localo)
    org_xyz_pngs = os.path.isfile(x_local) and os.path.isfile(y_local) and os.path.isfile(z_local)
    org_xyz_pngs_scaled = False
    if org_xyz_pngs:
        x_local_scaled = '{}_scaled_xplanes.png'.format(resmapname)
        y_local_scaled = '{}_scaled_yplanes.png'.format(resmapname)
        z_local_scaled = '{}_scaled_zplanes.png'.format(resmapname)
        size = (300,300)
        resize_apng(x_localo, x_local_scaled, size)
        resize_apng(y_localo, y_local_scaled, size)
        resize_apng(z_localo, z_local_scaled, size)
        org_xyz_pngs_scaled = os.path.isfile(x_local_scaled) and os.path.isfile(y_local_scaled) and \
                              os.path.isfile(z_local_scaled)

    if org_xyz_pngs and org_xyz_pngs_scaled:
        result_dict = {'local_resolution': {'ResMap': {'original': {'x': os.path.basename(x_local),
                                                                    'y': os.path.basename(y_local),
                                                                    'z': os.path.basename(z_local)}},
                                                        'scaled': {'x': os.path.basename(x_local_scaled),
                                                                   'y': os.path.basename(y_local_scaled),
                                                                   'z': os.path.basename(z_local_scaled)}}}
        try:
            with codecs.open(output_filename, 'w', encoding='utf-8') as f:
                json.dump(result_dict, f)
        except:
            sys.stderr.write(
                'Saving ResMap view to JSON error: {}.\n'.format(sys.exc_info()[1]))
    else:
        print('Missing ResMap result images.')


def make_apng_looping(input_file, output_file):
    """
        Chimerax give animated png for 1 loop need to change to 0 make a infinite loop
    :param input_file:
    :param output_file:
    :return:
    """

    image = Image.open(input_file)
    if image.info['loop'] != 0:
        image.info['loop'] = 0
    image.save(output_file, save_all=True)


def resize_apng(input_file, output_file, size):
    # Open the APNG file
    input_apng = Image.open(input_file)

    # Create a list to store resized frames
    resized_frames = []

    # Iterate over each frame of the APNG
    for frame in ImageSequence.Iterator(input_apng):
        # Resize the frame's image
        resized_image = frame.copy().resize(size, Image.Resampling.LANCZOS)
        resized_frames.append(resized_image)

    # Save the resized frames as an APNG
    resized_frames[0].save(output_file, save_all=True, append_images=resized_frames[1:], optimize=False,
                           duration=input_apng.info.get('duration', 100), loop=input_apng.info.get('loop', 0))


def get_masked_indices(input_mask):

    masked_map = mrcfile.open(input_mask)
    masked_map_data = masked_map.data > 0.5
    # masked_map_indices = np.argwhere(masked_map_data > 0.5)

    return masked_map_data


def localres_histogram(mapone, primary_mapname, resolution=None):
    """
        local resolution histogram
    """

    mapname = os.path.basename(mapone)
    output_path = os.path.dirname(mapone)
    file_pattern = '{}/{}_relion/*_locres.mrc'.format(output_path, primary_mapname)
    resmapname_glob = glob.glob(file_pattern)
    resmapname = resmapname_glob[0] if len(resmapname_glob) > 0 else None
    mask_file = f'{output_path}/{primary_mapname}_relion/mask/{primary_mapname}_mask.mrc'
    bin_mask = get_masked_indices(mask_file)
    final_dict = {}
    unmasked_localres = {}
    masked_localres = {}

    if resmapname and os.path.isfile(resmapname):
        with mrcfile.open(resmapname, permissive=True) as inmap:
            map_data = inmap.data[inmap.data != 0]
            tmasked_map_data = inmap.data*bin_mask
            masked_map_data = tmasked_map_data[tmasked_map_data != 0]
            hist, bin_edges = np.histogram(map_data, bins=50)
            masked_hist, masked_bin_edges = np.histogram(masked_map_data, bins=50)

            unmasked_localres['values'] = bin_edges.tolist()
            unmasked_localres['counts'] = hist.tolist()
            unmasked_localres['minmax'] = {'min': keep_three_significant_digits(min(bin_edges).astype(float)),
                                       'max': keep_three_significant_digits(max(bin_edges).astype(float))}
            masked_localres['values'] = masked_bin_edges.tolist()
            masked_localres['counts'] = masked_hist.tolist()
            masked_localres['minmax'] = {'min': keep_three_significant_digits(min(masked_bin_edges).astype(float)),
                                           'max': keep_three_significant_digits(max(masked_bin_edges).astype(float))}

            # Plot the histogram
            plt.figure(figsize=(8, 6))
            plt.hist(map_data, bins=50, color='green', edgecolor='black')
            plt.hist(masked_map_data, bins=50, color='orange', edgecolor='black')
            final_dict['masked'] = masked_localres
            final_dict['unmasked'] = unmasked_localres

            if resolution:
                vertical_lines = [resolution]  # Example vertical lines at x=3 and x=5
                for line in vertical_lines:
                    plt.axvline(x=line, color='red', linestyle='solid', linewidth=2)
                    plt.text(line, plt.ylim()[1] - 0.038 * (plt.ylim()[1] - plt.ylim()[0]), f' Claimed resolution={line} Å', color='red', va='bottom', ha='left')

            plt.ylabel('Voxel counts')
            plt.xlabel('Local resolution ')

            localres_histogram_plot = '{}/{}_localres_histogram.png'.format(output_path, primary_mapname)
            # Save the plot
            plt.savefig(localres_histogram_plot)

    return final_dict






import sys
import os
from TEMPy.maps.map_parser import MapParser
from TEMPy.protein.structure_parser import mmCIFParser
from TEMPy.protein.scoring_functions import FastSMOC
from collections import OrderedDict
from va.utils.misc import create_directory
from va.utils.misc import check_same_keys
from va.utils.misc import floatohex


def restructure_dict(data):
    """
    Reformat dictionary
    data: a dictionary as "('A', 659, 'PHE'), ('A', 416, 'GLU'), ('C', 651, 'GLU'), ('A', 498, 'GLN')}"
    return: a dictionary {'A': {416: 'GLU', 498: 'GLN', 659: 'PHE'}, 'C': {651: 'GLU'}}
    """

    restructured_dict = {}
    for item in data:
        chain, number, residue = item
        if chain not in restructured_dict:
            restructured_dict[chain] = {}
        restructured_dict[chain][number] = residue
    # Sort residue numbers within each chain
    for chain in restructured_dict:
        restructured_dict[chain] = dict(sorted(restructured_dict[chain].items()))
    return restructured_dict



def run_smoc(full_modelpath, full_mappath, res, output_path):
    """
        smoc score
    :return:
    """

    errlist = []
    result_dict = {}
    try:
        create_directory(output_path)
        cmap = MapParser.readMRC(full_mappath)
        model = mmCIFParser.read_mmCIF_file(full_modelpath)
        window = 11
        scorer = FastSMOC(model, cmap, float(res))
        chain_scores = {}
        chains = set(a.chain for a in model.atomList)
        residues_set = set((a.chain, a.res_no, a.res) for a in model.atomList)
        residues = restructure_dict(residues_set)
        for chain in chains:
            chain_scores[chain] = scorer.score_chain_contig(chain, window)
        smoc_result = [residues, chain_scores]

        model_name = os.path.basename(full_modelpath)
        result_dict = smoc_todict(model_name, smoc_result)
    except AssertionError:
        sys.stderr.write('SMOC does not work well. \n')

    return result_dict, errlist

def smoc_todict(model_name, smoc_result):
    """
        Save SMOC results into json file
    :param smoc_result:
    :return: json file name
    """

    colors = []
    smocs = []
    residues = []
    chain_smocs = OrderedDict()
    check = check_same_keys(smoc_result[0], smoc_result[1])
    if check:
        chain_all = smoc_result[0].keys()
        chain_length = len(chain_all)
        for chain in smoc_result[0].keys():
            chain_smoc = 0.
            residues_list = smoc_result[0][chain].keys()
            for residue_no in residues_list:
                residue_smoc = smoc_result[1][chain][residue_no]
                chain_smoc += residue_smoc
                residue_type = smoc_result[0][chain][residue_no]
                residue_color = floatohex([residue_smoc])[0]
                residue_string = '{}:{} {}'.format(chain, residue_no, residue_type)
                colors.append(residue_color)
                smocs.append(float('%.3f' % residue_smoc))
                residues.append(residue_string)

            chain_smocvalue = chain_smoc / chain_length if chain_length != 0. else 0.
            chain_smoccolor = floatohex([chain_smocvalue])[0]
            chain_smocs[chain] = {'value': float('%.3f' % chain_smocvalue), 'color': chain_smoccolor}

        average_smoc = float('%.3f' % (sum(smocs)/len(smocs)))
        average_smoc_color = floatohex([average_smoc])[0]

        data = {'averagesmoc': average_smoc, 'averagesmoc_color': average_smoc_color, 'color': colors,
                'smoc_scores': smocs, 'residue': residues, 'chainsmoc': chain_smocs}

        result_dict = {'name': model_name, 'data': data}

        return result_dict
    return None


import subprocess
import sys
import os
from distutils.spawn import find_executable


def create_folder(output_path):
    """
        create emringer output folder inside va directory

    :return: model related path of strudel folder
    """

    fullname = '{}'.format(output_path)

    if not os.path.isdir(fullname):
        os.mkdir(fullname, mode=0o777)
    else:
        print('{} is exist'.format(fullname))


def run_threedfsc(mapodd, mapeven, primarymap, vxsize, output_path, threedfscdir):
    """
        3DFSC calculation

    :return:
    """

    errlist = []
    try:
        assert find_executable('conda') is not None
        conda_path = find_executable('conda')
        if 'condabin' in conda_path:
            conda_root = conda_path.replace('/condabin/conda', '')
        else:
            conda_root = conda_path.replace('/bin/conda', '')
        source_conda = '{}/etc/profile.d/conda.sh'.format(conda_root)
        create_folder(output_path)
        conda_cmd = 'source {}; conda activate 3dfsc; python {} --halfmap1 {} --halfmap2 {} --fullmap {} --apix {}'.\
            format(source_conda, threedfscdir, mapodd, mapeven, primarymap, vxsize)
        try:
            subprocess.check_call(conda_cmd, cwd=output_path, shell=True)
        except subprocess.CalledProcessError as suberr:
            err = '3DFSC calculation error: {}.'.format(suberr)
            errlist.append(err)
            sys.stderr.write(err + '\n')
    except AssertionError:
        sys.stderr.write('conda executable is not there.\n')


    return errlist


